from typing import Optional, List

from ..util import prettify, get_optional, get_optional_list, get_required
from ..enums import GameModeStr
from .forum import TextFormat


__all__ = ("Group", "UserGroup")


class Group:
    """
    This object isn't returned by any endpoints yet, it is here purely as a reference for :class:`UserGroup`

    **Attributes**

    id: int

    identifier: str
        Unique string to identify the group.

    is_probationary: bool
        Whether members of this group are considered probationary.

    has_playmodes: bool
        If this group associates GameModes with a user's membership, e.g. BN/NAT members

    has_listing: bool
        Whether this group displays a listing at /groups/{id}

    name: str

    short_name: str
        Short name of the group for display.

    colour: Optional[str]

    description: Optional[:class:`TextFormat`]
        A dictionary with keys html and markdown.
    """

    __slots__ = (
        "id",
        "identifier",
        "is_probationary",
        "has_playmodes",
        "has_listing",
        "name",
        "short_name",
        "colour",
        "description",
    )

    def __init__(self, data):
        self.id: int = get_required(data, "id")
        self.identifier: str = get_required(data, "identifier")
        self.is_probationary: bool = get_required(data, "is_probationary")
        self.has_playmodes: bool = get_required(data, "has_playmodes")
        self.name: str = get_required(data, "name")
        self.short_name: str = get_required(data, "short_name")
        self.colour: Optional[str] = get_required(data, "colour")
        self.description: Optional[TextFormat] = get_optional(data, "description", TextFormat)

    def __repr__(self):
        return prettify(self, "name")


class UserGroup(Group):
    """
    Describes the :class:`Group` membership of a :class:`User`.
    It contains all of the attributes of the :class:`Group`, in addition to what is listed here.

    **Attributes**

    playmodes: Optional[List[:class:`GameModeStr`]]
        GameModes associated with this membership (None if `has_playmodes` is false).
    """

    __slots__ = ("playmodes",)

    def __init__(self, data):
        super().__init__(data)
        self.playmodes: Optional[List[GameModeStr]] = get_optional_list(data, "playmodes", GameModeStr)

    def __repr__(self):
        return super().__repr__()



from typing import Optional, List, TYPE_CHECKING

from ..util import prettify, get_optional, get_optional_list, get_required, fromisoformat
from ..enums import ChatChannelType, ChatMessageType
from .current_user_attributes import ChatChannelUserAttributes
from .user import UserCompact


if TYPE_CHECKING:
    from datetime import datetime


__all__ = ("ChatChannel", "ChatMessage")


class ChatChannel:
    """
    Represents an individual chat "channel" in the game.

    **Attributes**

    channel_id: int

    name: str

    description: Optional[str]

    icon: Optional[str]
        display icon for the channel

    type: :class:`ChatChannelType`

    moderated: bool
        user can't send message when the value is `True`

    uuid: Optional[str]
        value from requests that is relayed back to the sender.

    current_user_attributes: Optional[:class:`ChatChannelUserAttributes`]
        only present on some responses

    last_message_id: Optional[int]
        message_id of last known message (only returned in presence responses)

    recent_messages: Optional[List[:class:`ChatMessage`]]
        [DEPRECATED] up to 50 most recent messages

    users: Optional[List[int]]
        list of user ids that are in the channel (not included for PUBLIC channels).
    """

    __slots__ = (
        "channel_id",
        "name",
        "description",
        "icon",
        "type",
        "moderated",
        "uuid",
        "current_user_attributes",
        "last_message_id",
        "recent_messages",
        "users",
    )

    def __init__(self, data):
        self.channel_id: int = get_required(data, "channel_id")
        self.name: str = get_required(data, "name")
        self.description: Optional[str] = data.get("description")
        self.icon: Optional[str] = data.get("icon")
        self.type: ChatChannelType = ChatChannelType(get_required(data, "type"))
        self.moderated: bool = get_required(data, "moderated")

        self.uuid: Optional[str] = data.get("uuid")
        self.current_user_attributes: Optional[ChatChannelUserAttributes] = get_optional(
            data, "current_user_attributes", ChatChannelUserAttributes
        )
        self.last_message_id: Optional[int] = data.get("last_message_id")
        self.recent_messages: Optional[List[ChatMessage]] = get_optional_list(data, "recent_messages", ChatMessage)
        self.users: Optional[List[int]] = data.get("users")

    def __repr__(self):
        return prettify(self, "name")


class ChatMessage:
    """
    Represents an individual Message within a :class:`ChatChannel`.

    **Attributes**

    channel_id: int
        channel_id of where the message was sent

    content: str
        message content

    is_action: bool
        was this an action? i.e. /me dances

    message_id: int
        unique identifier for message

    sender_id: int
        user_id of the sender

    timestamp: :py:class:`datetime.datetime`
        when the message was sent

    type: :class:`ChatMessageType`

    uuid: Optional[str]
        message identifier originally sent by client

    sender: Optional[:class:`UserCompact`]
        embedded :class:`UserCompact` object to save additional api lookups
    """

    __slots__ = (
        "channel_id",
        "content",
        "is_action",
        "message_id",
        "sender_id",
        "timestamp",
        "type",
        "uuid",
        "sender",
    )

    def __init__(self, data):
        self.channel_id: int = get_required(data, "channel_id")
        self.content: str = get_required(data, "content")
        self.is_action: bool = get_required(data, "is_action")
        self.message_id: int = get_required(data, "message_id")
        self.sender_id: int = get_required(data, "sender_id")
        self.timestamp: datetime = fromisoformat(get_required(data, "timestamp"))
        self.type: ChatMessageType = ChatMessageType(get_required(data, "type"))
        self.uuid: Optional[str] = data.get("uuid")
        self.sender: Optional[UserCompact] = get_optional(data, "sender", UserCompact)

    def __repr__(self):
        return prettify(self, "sender_id" if self.sender is None else "sender", "content")



from .auth import *
from .client import *
from .exceptions import *
from .notification import NotificationWebsocket
from .asyncio import *
from .enums import *
from .objects import *
from .util import *
from .results import *
from .path import *
from .scope import *


__version__ = "4.1.0"



from typing import Optional, List, Union, TYPE_CHECKING

from ..enums import (
    MatchEventType,
    GameModeStr,
    GameModeInt,
    ScoringType,
    TeamType,
    Mods,
)
from ..util import prettify, get_optional, get_required, fromisoformat
from .user import UserCompact
from .beatmap import BeatmapCompact
from .score import LegacyScore, SoloScore, get_score_object


if TYPE_CHECKING:
    from datetime import datetime


__all__ = ("Match", "MatchExtended", "MatchEvent", "MatchGame", "MatchGameScoreInfo")


class Match:
    """
    Info of a match, relevant at :func:`osu.Client.get_matches`.

    **Attributes**

    id: int

    name: str

    start_time: Optional[:py:class:`datetime.datetime`]

    end_time: Optional[:py:class:`datetime.datetime`]
    """

    __slots__ = ("id", "start_time", "end_time", "name")

    def __init__(self, data):
        self.id: int = get_required(data, "id")
        self.name: str = get_required(data, "name")
        self.start_time: Optional[datetime] = get_optional(data, "start_time", fromisoformat)
        self.end_time: Optional[datetime] = get_optional(data, "end_time", fromisoformat)

    def __repr__(self):
        return prettify(self, "name", "start_time")


class MatchExtended(Match):
    """
    Extended version of :class:`Match` that is relevant at the :func:`osu.Client.get_match` endpoint.

    **Attributes**

    events: List[:class:`MatchEvent`]
        List of events that occurred in the match.

    users: List[:class:`UserCompact`]

    first_event_id: int

    latest_event_id: int

    current_game_id: int
    """

    __slots__ = (
        "events",
        "users",
        "first_event_id",
        "latest_event_id",
        "current_game_id",
    )

    def __init__(self, data, api_version):
        super().__init__(get_required(data, "match"))
        self.events: List[MatchEvent] = [MatchEvent(evt, api_version) for evt in get_required(data, "events")]
        self.users: List[UserCompact] = list(map(UserCompact, get_required(data, "users")))
        self.first_event_id: int = get_required(data, "first_event_id")
        self.latest_event_id: int = get_required(data, "latest_event_id")
        self.current_game_id: int = get_required(data, "current_game_id")


class MatchEvent:
    """
    An event that occurred in a match.

    **Attributes**

    id: int

    timestamp: :py:class:`datetime.datetime`

    user_id: int

    type: :class:`MatchEventType`

    text: Optional[str]
        None unless the event type is :class:`MatchEventType`.OTHER

    game: Optional[:class:`MatchGame`]
        None unless the event type is :class:`MatchEventType`.OTHER
    """

    __slots__ = ("id", "timestamp", "user_id", "type", "text", "game")

    def __init__(self, data, api_version):
        self.id: int = get_required(data, "id")
        self.timestamp: datetime = fromisoformat(get_required(data, "timestamp"))
        self.user_id: int = get_required(data, "user_id")
        self.type: MatchEventType = MatchEventType(get_required(data, "detail")["type"])
        self.text: Optional[str] = (
            get_required(data, "detail")["text"] if "text" in get_required(data, "detail") else None
        )
        self.game: Optional[MatchGame] = get_optional(data, "game", lambda d: MatchGame(d, api_version))

    def __repr__(self):
        attributes = ("type",) if self.type != MatchEventType.OTHER else ("type", "game")
        return prettify(self, *attributes)


class MatchGame:
    """
    Represents a map played in a match and contains all the info about the game

    **Attributes**

    beatmap_id: int

    id: int

    start_time: Optional[:py:class:`datetime.datetime`]

    end_time: Optional[:py:class:`datetime.datetime`]

    mode: :class:`GameModeStr`

    mode_int: :class:`GameModeInt`

    scoring_type: :class:`ScoringType`

    team_type: :class:`TeamType`

    mods: :class:`Mods`

    beatmap: Optional[:class:`BeatmapCompact`]

    scores: List[Union[:class:`LegacyScore`, :class:`SoloScore`]]
    """

    __slots__ = (
        "beatmap_id",
        "id",
        "start_time",
        "end_time",
        "mode",
        "mode_int",
        "scoring_type",
        "team_type",
        "mods",
        "beatmap",
        "scores",
    )

    def __init__(self, data, api_version):
        self.beatmap_id: int = get_required(data, "beatmap_id")
        self.id: int = get_required(data, "id")
        self.start_time: datetime = fromisoformat(get_required(data, "start_time"))
        self.end_time: Optional[datetime] = get_optional(data, "end_time", fromisoformat)
        self.mode: GameModeStr = GameModeStr(get_required(data, "mode"))
        self.mode_int: GameModeInt = GameModeInt(get_required(data, "mode_int"))
        self.scoring_type: ScoringType = ScoringType(get_required(data, "scoring_type"))
        self.team_type: TeamType = TeamType(get_required(data, "team_type"))
        self.mods: Mods = Mods.parse_any_list(get_required(data, "mods"))
        self.beatmap: Optional[BeatmapCompact] = get_optional(data, "beatmap", BeatmapCompact)
        self.scores: List[Union[LegacyScore, SoloScore]] = [
            get_score_object(score, api_version) for score in get_required(data, "scores")
        ]

    def __repr__(self):
        return prettify(self, "beatmap", "scores")


class MatchGameScoreInfo:
    """
    Says info about a score set in a match. Is an attribute of :class:`LegacyScore`

    **Attributes**

    slot: int

    team: str

    passed: bool
    """

    __slots__ = ("slot", "team", "passed")

    def __init__(self, data):
        self.slot: int = get_required(data, "slot")
        self.team: str = get_required(data, "team")
        self.passed: bool = get_required(data, "pass")

    def __repr__(self):
        return prettify(self, "slot", "team")



from typing import Optional, TYPE_CHECKING, Union

from ..util import prettify, get_required, fromisoformat
from ..enums import GameModeStr, RankStatus
from .achievement import Achievement


if TYPE_CHECKING:
    from datetime import datetime


__all__ = (
    "Event",
    "AchievementEvent",
    "BeatmapPlaycountEvent",
    "BeatmapsetApproveEvent",
    "BeatmapsetDeleteEvent",
    "BeatmapsetReviveEvent",
    "BeatmapsetUpdateEvent",
    "BeatmapsetUploadEvent",
    "RankEvent",
    "RankLostEvent",
    "UserSupportAgainEvent",
    "UserSupportFirstEvent",
    "UserSupportGiftEvent",
    "UsernameChangeEvent",
    "EventUser",
    "EventBeatmap",
    "EventBeatmapset",
    "EVENT_TYPE",
    "get_event_object",
)


class Event:
    """
    Base class of an event

    **Attributes**

    created_at: :py:class:`datetime.datetime`

    id: int

    type:
        All types and the additional attributes they provide are listed under 'Event Types'
    """

    def __init__(self, data):
        self.created_at: datetime = fromisoformat(get_required(data, "created_at"))
        self.id: int = get_required(data, "id")


class AchievementEvent(Event):
    """
    **Attributes**

    achievement: :class:`Achievement`

    user: :class:`EventUser`
    """

    __slots__ = ("achievement", "user")

    def __init__(self, data):
        super().__init__(data)
        self.achievement: Achievement = Achievement(get_required(data, "achievement"))
        self.user: EventUser = EventUser(get_required(data, "user"))

    def __repr__(self):
        return prettify(self, "achievement", "user")


class BeatmapPlaycountEvent(Event):
    """
    **Attributes**

    beatmap: :class:`EventBeatmap`

    count: int
    """

    __slots__ = ("beatmap", "count")

    def __init__(self, data):
        super().__init__(data)
        self.beatmap: EventBeatmap = EventBeatmap(get_required(data, "beatmap"))
        self.count: int = get_required(data, "count")

    def __repr__(self):
        return prettify(self, "beatmap", "count")


class BeatmapsetApproveEvent(Event):
    """
    **Attributes**

    approval: :class:`RankStatus`

    beatmapset: :class:`EventBeatmapset`

    user: :class:`EventUser`
    """

    __slots__ = ("approval", "beatmapset", "user")

    def __init__(self, data):
        super().__init__(data)
        self.approval: RankStatus = RankStatus[get_required(data, "approval").upper()]
        self.beatmapset: EventBeatmapset = EventBeatmapset(get_required(data, "beatmapset"))
        self.user: EventUser = EventUser(get_required(data, "user"))

    def __repr__(self):
        return prettify(self, "beatmapset", "approval")


class BeatmapsetDeleteEvent(Event):
    """
    **Attributes**

    beatmapset: :class:`EventBeatmapset`
    """

    __slots__ = ("beatmapset",)

    def __init__(self, data):
        super().__init__(data)
        self.beatmapset: EventBeatmapset = EventBeatmapset(get_required(data, "beatmapset"))

    def __repr__(self):
        return prettify(self, "beatmapset")


class BeatmapsetReviveEvent(Event):
    """
    **Attributes**

    beatmapset: :class:`EventBeatmapset`

    user: :class:`EventUser`
    """

    __slots__ = ("beatmapset", "user")

    def __init__(self, data):
        super().__init__(data)
        self.beatmapset: EventBeatmapset = EventBeatmapset(get_required(data, "beatmapset"))
        self.user: EventUser = EventUser(get_required(data, "user"))

    def __repr__(self):
        return prettify(self, "beatmapset", "user")


class BeatmapsetUpdateEvent(Event):
    """
    **Attributes**

    beatmapset: :class:`EventBeatmapset`

    user: :class:`EventUser`
    """

    __slots__ = ("beatmapset", "user")

    def __init__(self, data):
        super().__init__(data)
        self.beatmapset: EventBeatmapset = EventBeatmapset(get_required(data, "beatmapset"))
        self.user: EventUser = EventUser(get_required(data, "user"))

    def __repr__(self):
        return prettify(self, "beatmapset", "user")


class BeatmapsetUploadEvent(Event):
    """
    **Attributes**

    beatmapset: :class:`EventBeatmapset`

    user: :class:`EventUser`
    """

    __slots__ = ("beatmapset", "user")

    def __init__(self, data):
        super().__init__(data)
        self.beatmapset: EventBeatmapset = EventBeatmapset(get_required(data, "beatmapset"))
        self.user: EventUser = EventUser(get_required(data, "user"))

    def __repr__(self):
        return prettify(self, "beatmapset", "user")


class RankEvent(Event):
    """
    **Attributes**

    score_rank: str

    rank: int

    mode: :class:`GameModeStr`

    beatmap: :class:`EventBeatmap`

    user: :class:`EventUser`
    """

    __slots__ = ("score_rank", "rank", "mode", "beatmap", "user")

    def __init__(self, data):
        super().__init__(data)
        self.score_rank: str = get_required(data, "scoreRank")
        self.rank: int = get_required(data, "rank")
        self.mode: GameModeStr = GameModeStr(get_required(data, "mode"))
        self.beatmap: EventBeatmap = EventBeatmap(get_required(data, "beatmap"))
        self.user: EventUser = EventUser(get_required(data, "user"))

    def __repr__(self):
        return prettify(self, "rank", "beatmap")


class RankLostEvent(Event):
    """
    **Attributes**

    mode: :class:`GameModeStr`

    beatmap: :class:`EventBeatmap`

    user: :class:`EventUser`
    """

    __slots__ = ("mode", "beatmap", "user")

    def __init__(self, data):
        super().__init__(data)
        self.mode: GameModeStr = GameModeStr(get_required(data, "mode"))
        self.beatmap: EventBeatmap = EventBeatmap(get_required(data, "beatmap"))
        self.user: EventUser = EventUser(get_required(data, "user"))

    def __repr__(self):
        return prettify(self, "beatmap", "user")


class UserSupportAgainEvent(Event):
    """
    **Attributes**

    user: :class:`EventUser`
    """

    __slots__ = ("user",)

    def __init__(self, data):
        super().__init__(data)
        self.user: EventUser = EventUser(get_required(data, "user"))

    def __repr__(self):
        return prettify(self, "user")


class UserSupportFirstEvent(Event):
    """
    **Attributes**

    user: :class:`EventUser`
    """

    __slots__ = ("user",)

    def __init__(self, data):
        super().__init__(data)
        self.user: EventUser = EventUser(get_required(data, "user"))

    def __repr__(self):
        return prettify(self, "user")


class UserSupportGiftEvent(Event):
    """
    **Attributes**

    user: :class:`EventUser`
    """

    __slots__ = ("user",)

    def __init__(self, data):
        super().__init__(data)
        self.user: EventUser = EventUser(get_required(data, "user"))

    def __repr__(self):
        return prettify(self, "user")


class UsernameChangeEvent(Event):
    """
    **Attributes**

    user: :class:`EventUser`
    """

    __slots__ = ("user",)

    def __init__(self, data):
        super().__init__(data)
        self.user: EventUser = EventUser(get_required(data, "user"))

    def __repr__(self):
        return prettify(self, "user")


class EventUser:
    """
    **Attributes**

    username: str

    url: str

    previous_username: Optional[str]
        Only for UsernameChangeEvent.
    """

    __slots__ = ("username", "url", "previous_username")

    def __init__(self, data):
        self.username: str = get_required(data, "username")
        self.url: str = get_required(data, "url")
        self.previous_username: Optional[str] = data.get("previousUsername")

    def __repr__(self):
        return prettify(self, "username")


class EventBeatmap:
    """
    **Attributes**

    title: str

    url: str
    """

    __slots__ = ("title", "url")

    def __init__(self, data):
        self.title: str = get_required(data, "title")
        self.url: str = get_required(data, "url")

    def __repr__(self):
        return prettify(self, "title")


class EventBeatmapset:
    """
    **Attributes**

    title: str

    url: str
    """

    __slots__ = ("title", "url")

    def __init__(self, data):
        self.title: str = get_required(data, "title")
        self.url: str = get_required(data, "url")

    def __repr__(self):
        return prettify(self, "title")


EVENT_TYPE = Union[
    AchievementEvent,
    BeatmapPlaycountEvent,
    BeatmapsetApproveEvent,
    BeatmapsetDeleteEvent,
    BeatmapsetReviveEvent,
    BeatmapsetUpdateEvent,
    BeatmapsetUploadEvent,
    RankEvent,
    RankLostEvent,
    UserSupportAgainEvent,
    UserSupportFirstEvent,
    UserSupportGiftEvent,
    UsernameChangeEvent,
]


def get_event_object(data) -> EVENT_TYPE:
    t = get_required(data, "type")
    cls = globals().get(t[0].upper() + t[1:] + "Event", data)
    return cls(data)



from typing import Optional, TYPE_CHECKING

from ..util import prettify, get_optional, get_required, fromisoformat
from ..enums import KudosuAction, ObjectType


if TYPE_CHECKING:
    from datetime import datetime


__all__ = ("KudosuHistory", "KudosuPost", "KudosuGiver")


class KudosuHistory:
    """
    **Attributes**

    id: int

    action: :class:`KudosuAction`

    amount: int

    model: :class:`ObjectType`
        Object type which the exchange happened on (forum_post, etc).

    created_at: :py:class:`datetime.datetime`

    giver: Optional[:class:`KudosuGiver`]
        Simple detail of the user who started the exchange.

    post: :class:`KudosuPost`
        Simple detail of the object for display.
    """

    __slots__ = ("id", "action", "amount", "model", "created_at", "giver", "post")

    def __init__(self, data):
        self.id: int = get_required(data, "id")
        self.action: KudosuAction = KudosuAction(get_required(data, "action"))
        self.amount: int = get_required(data, "amount")
        self.model: ObjectType = ObjectType(get_required(data, "model"))
        self.created_at: datetime = fromisoformat(get_required(data, "created_at"))
        self.giver: Optional[KudosuGiver] = get_optional(data, "giver", KudosuGiver)
        self.post: KudosuPost = KudosuPost(get_required(data, "post"))

    def __repr__(self):
        return prettify(self, "action", "amount", "giver")


class KudosuPost:
    """
    **Attributes**

    url: Optional[str]
        Url of the object.

    title: str
        Title of the object. It'll be "[deleted beatmap]" for deleted beatmaps.
    """

    __slots__ = ("url", "title")

    def __init__(self, data):
        self.url: Optional[str] = get_required(data, "url")
        self.title: str = get_required(data, "title")

    def __repr__(self):
        return prettify(self, "title")


class KudosuGiver:
    """
    **Attributes**

    url: str

    username: str
    """

    __slots__ = ("url", "username")

    def __init__(self, data):
        self.url: str = get_required(data, "url")
        self.username: str = get_required(data, "username")

    def __repr__(self):
        return prettify(self, "username")



from dataclasses import dataclass
from typing import Dict, Optional, List, TypeVar, Type, TYPE_CHECKING

from .objects import (
    Beatmap,
    Beatmapset,
    BeatmapsetCompact,
    BeatmapsetDiscussion,
    BeatmapsetDiscussionPost,
    BeatmapsetDiscussionVote,
    BeatmapsetEvent,
    Build,
    ChatChannel,
    ChatMessage,
    Forum,
    ForumPost,
    ForumTopic,
    Match,
    NewsPost,
    Review,
    SoloScore,
    UpdateStream,
    UserCompact,
    UserScoreAggregate,
    WikiPage,
)

_T = TypeVar("_T")


def result_dataclass(cls: Type[_T]) -> Type[_T]:
    """basic wrapper that adds slots"""
    from dataclasses import dataclass as _dataclass, fields

    cls = _dataclass(frozen=True)(cls)

    cls_dict = dict(cls.__dict__)
    field_names = tuple(field.name for field in fields(cls))
    cls_dict["__slots__"] = field_names
    for field_name in field_names:
        cls_dict.pop(field_name, None)
    cls_dict.pop("__dict__", None)

    return type(cls)(cls.__name__, cls.__bases__, cls_dict)


# "trick" type checkers into seeing the classes as dataclasses
# the type checkers can't tell that they are otherwise
if not TYPE_CHECKING:
    dataclass = result_dataclass
    del result_dataclass

__all__ = (
    "BeatmapsetSearchResult",
    "BeatmapsetDiscussionPostsResult",
    "BeatmapsetDiscussionVotesResult",
    "ReviewsConfig",
    "BeatmapsetDiscussionsResult",
    "ChangelogListingSearch",
    "ChangelogListingResult",
    "CreateNewPmResult",
    "CreateTopicResult",
    "GetTopicAndPostsResult",
    "SearchResult",
    "SearchInfo",
    "GetNewsListingResult",
    "NewsSidebar",
    "GetBeatmapsetEventsResult",
    "GetMatchesResult",
    "GetRoomLeaderboardResult",
    "GetChannelResult",
    "GetAllScoresResult",
    "GetForumResult",
    "GetForumsResult",
    "GetForumTopicsResult",
)


class ResultBase:
    __slots__ = ()

    # For backwards compatibility
    def __getitem__(self, item):
        if isinstance(item, str):
            return getattr(self, item)
        elif isinstance(item, int):
            return self.__slots__[item], getattr(self, self.__slots__[item], None)
        raise AttributeError(f"Could not fetch attribute of {self.__class__.__name__} " f"from item value {item!r}")


@dataclass
class SearchInfo(ResultBase):
    """
    A class for search info in several result objects.

    **Attributes**

    sort: str

    limit: int

    start: Optional[str]

    end: Optional[str]
    """

    sort: str
    limit: int
    start: Optional[int]
    end: Optional[int]


@dataclass
class BeatmapsetDiscussionPostsResult(ResultBase):
    """
    Result of :func:`osu.Client.get_beatmapset_discussion_posts`

    **Attributes**

    beatmapsets: List[:class:`BeatmapsetCompact`]

    posts: List[:class:`BeatmapsetDiscussionPost`]

    users: List[:class:`UserCompact`]

    cursor: Dict[str, int]
    """

    beatmapsets: List[BeatmapsetCompact]
    posts: List[BeatmapsetDiscussionPost]
    users: List[UserCompact]
    cursor: str


@dataclass
class BeatmapsetDiscussionVotesResult(ResultBase):
    """
    Result of :func:`osu.Client.get_beatmapset_discussion_votes`

    **Attributes**

    discussions: List[:class:`BeatmapsetDiscussion`]

    votes: List[:class:`BeatmapsetDiscussionVote`]

    users: List[:class:`UserCompact`]

    cursor: Dict[str, int]
    """

    discussions: List[BeatmapsetDiscussion]
    votes: List[BeatmapsetDiscussionVote]
    users: List[UserCompact]
    cursor: str


@dataclass
class ReviewsConfig(ResultBase):
    """
    An attribute of :class:`BeatmapsetDiscussionsResult`

    **Attributes**

    max_blocks: int
    """

    max_blocks: int


@dataclass
class BeatmapsetDiscussionsResult(ResultBase):
    """
    Result of :func:`osu.Client.get_beatmapset_discussions`

    **Attributes**

    beatmaps: List[:class:`Beatmap`]

    discussions: List[:class:`BeatmapsetDiscussion`]

    included_discussions: List[:class:`BeatmapsetDiscussion`]

    users: List[:class:`UserCompact`]

    reviews_config: :class:`ReviewsConfig`

    cursor: str
    """

    beatmaps: List[Beatmap]
    discussions: List[BeatmapsetDiscussion]
    included_discussions: List[BeatmapsetDiscussion]
    users: List[UserCompact]
    reviews_config: ReviewsConfig
    cursor: str


@dataclass
class ChangelogListingSearch(ResultBase):
    """
    Attribute of :class:`ChangelogListingResult`

    **Attributes**

    start: Optional[str]
        `start` input

    end: Optional[str]
        `end` input

    limit: int
        Always 21

    max_id: Optional[int]
        `max_id` input

    stream: Optional[str]
        `stream` input
    """

    start: Optional[str]
    end: Optional[str]
    limit: int
    max_id: Optional[int]
    stream: Optional[str]


@dataclass
class ChangelogListingResult(ResultBase):
    """
    Result of :func:`osu.Client.get_changelog_listing`

    **Attributes**

    build: List[:class:`Build`]

    streams: List[:class:`Build`]

    search: :class:`ChangelogListingSearch`
    """

    builds: List[Build]
    streams: List[UpdateStream]
    search: ChangelogListingSearch


@dataclass
class CreateNewPmResult(ResultBase):
    """
    Result of :func:`osu.Client.create_new_pm`

    **Attributes**

    channel: :class:`ChatChannel`

    message: :class:`ChatMessage`

    new_channel_id: int
        [DEPRECATED] channel id of newly created :class:`ChatChannel`
    """

    channel: ChatChannel
    message: ChatMessage
    new_channel_id: int


@dataclass
class CreateTopicResult(ResultBase):
    """
    Result of :func:`osu.Client.create_topic`

    **Attributes**

    topic: :class:`ForumTopic`

    post: :class:`ForumPost`
        includes body
    """

    topic: ForumTopic
    post: ForumPost


@dataclass
class GetTopicAndPostsResult(ResultBase):
    """
    Result of :func:`osu.Client.get_topic_and_posts`

    **Attributes**

    cursor_string: str

    search: :class:`GetTopicAndPostsSearch`
        Parameters used for current request excluding cursor.

    topic: :class:`ForumTopic`

    posts: List[:class:`ForumPost`]
    """

    cursor_string: str
    search: SearchInfo
    topic: ForumTopic
    posts: List[ForumPost]


@dataclass
class SearchResult(ResultBase):
    """
    Result of :func:`osu.Client.search`

    **Attributes**

    user: Optional[List[:class:`UserCompact`]]
        For `all` or `user` mode. Only first 100 results are accessible

    wiki_page: Optional[List[:class:`WikiPage`]]
        For `all` or `wiki_page` mode
    """

    user: Optional[List[UserCompact]]
    wiki_page: Optional[List[WikiPage]]


@dataclass
class NewsSidebar(ResultBase):
    """
    Attribute of :class:`GetNewsListingResult`

    **Attributes**

    current_year: int

    news_post: List[:class:`NewsPost`]

    years: List[int]
    """

    current_year: int
    news_post: List[NewsPost]
    years: List[int]


@dataclass
class GetNewsListingResult(ResultBase):
    """
    Result of :func:`osu.Client.get_news_listing`

    **Attributes**

    cursor: str

    news_posts: List[:class:`NewsPost`]

    news_sidebar: :class:`NewsSidebar`

    search: :class:`SearchInfo`
    """

    cursor: str
    news_posts: List[NewsPost]
    news_sidebar: NewsSidebar
    search: SearchInfo


@dataclass
class GetBeatmapsetEventsResult(ResultBase):
    """
    Result of :func:`osu.Client.get_beatmapset_events`

    **Attributes**

    events: List[:class:`BeatmapsetEvent`]

    reviews_config: :class:`Review`

    users: List[:class:`UserCompact`]
        Includes `groups` attribute
    """

    events: List[BeatmapsetEvent]
    reviews_config: Review
    users: List[UserCompact]


@dataclass
class GetMatchesResult(ResultBase):
    """
    Result of :func:`osu.Client.get_matches`

    **Attributes**

    matches: List[:class:`Match`]

    params: Dict

    cursor: Dict
    """

    matches: List[Match]
    params: Dict
    cursor: Dict


@dataclass
class BeatmapsetSearchResult(ResultBase):
    """
    Result of :func:`osu.Client.search_beatmapsets`

    **Attributes**

    beatmapsets: List[:class:`Beatmapset`]

    cursor: Dict

    search: :class:`SearchInfo`

    recommended_difficulty: Optional[int]

    error: Optional[str]

    total: int
    """

    beatmapsets: List[Beatmapset]
    cursor: Dict
    search: SearchInfo
    recommended_difficulty: Optional[float]
    error: Optional[str]
    total: int


@dataclass
class GetRoomLeaderboardResult(ResultBase):
    """
    Result of :func:`osu.Client.get_room_leaderboard`

    **Attributes**

    leaderboard: List[:class:`UserScoreAggregate`]

    user_score: Optional[:class:`UserScoreAggregate`]
    """

    leaderboard: List[UserScoreAggregate]
    user_score: Optional[UserScoreAggregate]


@dataclass
class GetChannelResult(ResultBase):
    """
    Result of :func:`osu.Client.get_channel`

    **Attributes**

    channel: :class:`ChatChannel`

    users: List[:class:`UserCompact`]
    """

    channel: ChatChannel
    users: List[UserCompact]


@dataclass
class GetAllScoresResult(ResultBase):
    """
    Result of :func:`osu.Client.get_all_scores`

    **Attributes**

    scores: List[:class:`SoloScore`]

    cursor: str
    """

    scores: List[SoloScore]
    cursor: str


@dataclass
class GetForumsResult(ResultBase):
    """
    Result of :func:`osu.Client.get_forums`

    **Attributes**

    forums: List[:class:`Forum`]
    """

    forums: List[Forum]


@dataclass
class GetForumResult(ResultBase):
    """
    Result of :func:`osu.Client.get_forum`

    **Attributes**

    forum: :class:`Forum`

    topics: List[:class:`ForumTopic`]

    pinned_topics: List[:class:`ForumTopic`]
    """

    forum: Forum
    topics: List[ForumTopic]
    pinned_topics: List[ForumTopic]


@dataclass
class GetForumTopicsResult(ResultBase):
    """
    Result of :func:`osu.Client.get_forum_topics`

    **Attributes**

    topics: List[:class:`ForumTopic`]

    cursor: Optional[str]
    """

    topics: List[ForumTopic]
    cursor: Optional[str]



from typing import Optional, List, TYPE_CHECKING, Union

from ..util import prettify, get_optional, get_optional_list, get_required, fromisoformat
from ..enums import MessageType
from .current_user_attributes import BeatmapsetDiscussionPermissions
from .beatmap import BeatmapCompact, BeatmapsetCompact


if TYPE_CHECKING:
    from datetime import datetime


__all__ = (
    "BeatmapsetDiscussion",
    "VotesSummary",
    "VotersSummary",
    "BeatmapsetDiscussionPost",
    "SystemDiscussionPostMessage",
    "BeatmapsetDiscussionVote",
    "Review",
)


class BeatmapsetDiscussion:
    """
    Represents a Beatmapset modding discussion

    **Attributes**

    beatmap_id: Optional[int]

    beatmapset_id: int

    can_be_resolved: bool

    can_grant_kudosu: bool

    created_at: :py:class:`datetime.datetime`

    deleted_at: Optional[:py:class:`datetime.datetime`]

    deleted_by_id: Optional[int]

    id: int

    kudosu_denied: bool

    last_post_at: :py:class:`datetime.datetime`

    message_type: :class:`MessageType`

    parent_id: Optional[int]

    resolved: bool

    timestamp: Optional[int]

    updated_at: :py:class:`datetime.datetime`

    user_id: int

    beatmap: Optional[:class:`BeatmapCompact`]

    beatmapset: Optional[:class:`BeatmapsetCompact`]

    current_user_attributes: Optional[:class:`BeatmapsetDiscussionPermissions`]

    posts: Optional[List[:class:`BeatmapsetDiscussionPost`]]

    starting_post: Optional[:class:`BeatmapsetDiscussionPost`]

    votes: Optional[:class:`VotesSummary`]
    """

    __slots__ = (
        "beatmap_id",
        "beatmapset_id",
        "can_be_resolved",
        "can_grant_kudosu",
        "created_at",
        "deleted_at",
        "deleted_by_id",
        "id",
        "kudosu_denied",
        "last_post_at",
        "message_type",
        "parent_id",
        "resolved",
        "timestamp",
        "updated_at",
        "user_id",
        "beatmap",
        "beatmapset",
        "current_user_attributes",
        "posts",
        "starting_post",
        "votes",
    )

    def __init__(self, data):
        self.beatmap_id: Optional[int] = get_required(data, "beatmap_id")
        self.beatmapset_id: int = get_required(data, "beatmapset_id")
        self.can_be_resolved: bool = get_required(data, "can_be_resolved")
        self.can_grant_kudosu: bool = get_required(data, "can_grant_kudosu")
        self.created_at: datetime = fromisoformat(get_required(data, "created_at"))
        self.deleted_at: Optional[datetime] = get_optional(data, "deleted_at", fromisoformat)
        self.deleted_by_id: Optional[int] = get_required(data, "deleted_by_id")
        self.id: int = get_required(data, "id")
        self.kudosu_denied: bool = get_required(data, "kudosu_denied")
        self.last_post_at: datetime = fromisoformat(get_required(data, "last_post_at"))
        self.message_type: MessageType = MessageType(get_required(data, "message_type"))
        self.parent_id: Optional[int] = get_required(data, "parent_id")
        self.resolved: bool = get_required(data, "resolved")
        self.timestamp: Optional[int] = data.get("timestamp")
        self.updated_at: datetime = fromisoformat(get_required(data, "updated_at"))
        self.user_id: int = get_required(data, "user_id")

        self.beatmap: Optional[BeatmapCompact] = get_optional(data, "beatmap", BeatmapCompact)
        self.beatmapset: Optional[BeatmapsetCompact] = get_optional(data, "beatmapset", BeatmapsetCompact)
        self.current_user_attributes: Optional[BeatmapsetDiscussionPermissions] = get_optional(
            data, "current_user_attributes", BeatmapsetDiscussionPermissions
        )
        self.posts: Optional[List[BeatmapsetDiscussionPost]] = get_optional_list(
            data, "posts", BeatmapsetDiscussionPost
        )
        self.starting_post: Optional[BeatmapsetDiscussionPost] = get_optional(
            data, "starting_post", BeatmapsetDiscussionPost
        )
        self.votes: VotesSummary = get_optional(data, "votes", VotesSummary)

    def __repr__(self):
        return prettify(self, "beatmapset", "starting_post")


class VotesSummary:
    """
    Summarizes the votes of a discussion

    **Attributes**

    down: int
        Number of downvotes

    up: int
        Number of upvotes

    voters: :class:`VotersSummary`
        summarizes who voted up and down
    """

    __slots__ = ("down", "up", "voters")

    def __init__(self, data):
        self.down: int = get_required(data, "down")
        self.up: int = get_required(data, "up")
        self.voters: VotersSummary = VotersSummary(get_required(data, "voters"))


class VotersSummary:
    """
    Gives a summary of players who voted up and down on a discussion

    **Attributes**

    down: List[int]
        List of user IDs of users that downvoted

    up: List[int]
        List of user IDs of users that upvoted
    """

    __slots__ = ("down", "up")

    def __init__(self, data):
        self.down: List[int] = get_required(data, "down")
        self.up: List[int] = get_required(data, "up")


class BeatmapsetDiscussionPost:
    """
    Represents a post in a :class:`BeatmapsetDiscussion`.

    **Attributes**

    beatmapset_discussion_id: int

    created_at: :py:class:`datetime.datetime`

    deleted_at: Optional[:py:class:`datetime.datetime`]

    deleted_by_id: Optional[int]

    id: int

    last_editor_id: Optional[int]

    updated_at: :py:class:`datetime.datetime`

    user_id: int

    message: Union[str, :class:`SystemDiscussionPostMessage`]
        String when `system` is false and :class:`SystemDiscussionPostMessage` otherwise.

    system: bool

    beatmap_discussion: Optional[:class:`BeatmapsetDiscussion`]
    """

    __slots__ = (
        "beatmapset_discussion_id",
        "created_at",
        "deleted_at",
        "deleted_by_id",
        "id",
        "last_editor_id",
        "updated_at",
        "user_id",
        "message",
        "system",
        "beatmap_discussion",
    )

    def __init__(self, data):
        self.beatmapset_discussion_id: int = get_required(data, "beatmapset_discussion_id")
        self.created_at: datetime = fromisoformat(get_required(data, "created_at"))
        self.deleted_at: Optional[datetime] = get_optional(data, "deleted_at", fromisoformat)
        self.deleted_by_id: Optional[int] = data.get("deleted_by_id")
        self.id: int = get_required(data, "id")
        self.last_editor_id: Optional[int] = data.get("last_editor_id")
        self.updated_at: datetime = fromisoformat(get_required(data, "updated_at"))
        self.user_id: int = get_required(data, "user_id")

        self.system: bool = get_required(data, "system")
        self.message: Union[str, SystemDiscussionPostMessage] = (
            SystemDiscussionPostMessage(get_required(data, "message")) if self.system else get_required(data, "message")
        )

        self.beatmap_discussion: Optional[BeatmapsetDiscussion] = get_optional(
            data, "beatmap_discussion", BeatmapsetDiscussion
        )

    def __repr__(self):
        return prettify(self, "user_id", "message")


class SystemDiscussionPostMessage:
    """
    System message of :class:`BeatmapsetDiscussionPost`

    **Attributes**

    type: str

    value: bool
    """

    __slots__ = ("type", "value")

    def __init__(self, data):
        self.type: str = get_required(data, "type")
        self.value: bool = get_required(data, "value")


class BeatmapsetDiscussionVote:
    """
    Represents a vote on a :class:`BeatmapsetDiscussion`.

    **Attributes**

    beatmapset_discussion_id: int

    created_at: :py:class:`datetime.datetime`

    id: int

    score: int

    updated_at: :py:class:`datetime.datetime`

    user_id: int
    """

    __slots__ = (
        "beatmapset_discussion_id",
        "created_at",
        "id",
        "score",
        "updated_at",
        "user_id",
    )

    def __init__(self, data):
        self.beatmapset_discussion_id: int = get_required(data, "beatmapset_discussion_id")
        self.created_at: datetime = fromisoformat(get_required(data, "created_at"))
        self.id: int = get_required(data, "id")
        self.score: int = get_required(data, "score")
        self.updated_at: datetime = fromisoformat(get_required(data, "updated_at"))
        self.user_id: int = get_required(data, "user_id")

    def __repr__(self):
        return prettify(self, "user_id", "score")


class Review:
    """
    **Attributes**

    max_blocks: int
    """

    __slots__ = ("max_blocks",)

    def __init__(self, data):
        self.max_blocks: int = get_required(data, "max_blocks")



import requests
import time
import threading
import logging
from datetime import datetime, timezone
from typing import Optional, List, TYPE_CHECKING

from .exceptions import ScopeException, RequestException
from .constants import (
    DEFAULT_BASE_URL,
    DEFAULT_AUTH_URL,
    DEFAULT_TOKEN_URL,
    DEFAULT_DOMAIN,
    auth_url,
    token_url,
    base_url,
)
from .path import Path

if TYPE_CHECKING:
    from .auth import BaseAuthHandler
    from .asyncio.http import AsynchronousHTTPHandler
    from .asyncio.auth import BaseAsynchronousAuthHandler


__all__ = ("BaseHTTPHandler", "HTTPHandler")


_log = logging.getLogger(__name__)


def _convert_param_value(value):
    if isinstance(value, bool):
        return "true" if value else "false"
    return value


class BaseHTTPHandler:
    __slots__ = ("auth", "api_version", "domain", "base_url", "auth_url", "token_url")

    DEFAULT_API_VERSION = "20251007"

    def __init__(self, auth: Optional["BaseAuthHandler"], api_version: Optional[str] = None):
        self.auth: Optional[BaseAuthHandler] = auth
        self.set_api_version(api_version)

        self.domain = DEFAULT_DOMAIN
        self.auth_url = DEFAULT_AUTH_URL
        self.token_url = DEFAULT_TOKEN_URL
        self.base_url = DEFAULT_BASE_URL

    def set_domain(self, domain: str) -> None:
        self.domain = domain
        self.auth_url = auth_url(domain)
        self.token_url = token_url(domain)
        self.base_url = base_url(domain)

    def set_api_version(self, api_version: Optional[str]):
        self.api_version: str = api_version or self.DEFAULT_API_VERSION

    def set_ratelimit(self, request_wait_time: float = 1.0, limit_per_minute: int = 60):
        raise NotImplementedError()

    def check_path_validity(self, path: Path):
        if path.requires_auth and self.auth is None:
            raise ScopeException("You need to be authenticated to make this request.")

        if path.requires_auth and path.scope not in self.auth.scope:
            raise ScopeException(f"You don't have the {path.scope} scope, which is required to make this request.")

        if path.requires_user and not self.auth.has_user():
            raise ScopeException(
                "This request requires a user. You need either a delegate scope or "
                "to register OAuth with Authorization Code Grant."
            )

    def make_request(self, path, *args, **kwargs):
        raise NotImplementedError()


class HTTPHandler(BaseHTTPHandler):
    """
    Handles making requests. Used by :class:`osu.Client`.
    """

    __slots__ = ("rate_limit",)

    def __init__(
        self,
        auth: Optional["BaseAuthHandler"],
        request_wait_time: float = 1.0,
        limit_per_minute: int = 60,
        api_version: Optional[str] = None,
    ):
        super().__init__(auth, api_version)

        self.rate_limit: RateLimitHandler = RateLimitHandler(request_wait_time, limit_per_minute)

    def set_ratelimit(self, request_wait_time: float = 1.0, limit_per_minute: int = 60):
        self.rate_limit.wait_time = request_wait_time
        self.rate_limit.limit = limit_per_minute

    def get_headers(self, path, is_files=False, **kwargs):
        headers = {
            "charset": "utf-8",
            "x-api-version": self.api_version,
            "Accept": path.accept,
        }
        if not is_files:  # otherwise let requests library handle it
            headers["Content-Type"] = path.content_type

        if path.requires_auth and "Authorization" not in headers:
            token = self.auth.get_token()
            if token is None:
                raise ValueError("Cannot make request requiring authorization with a null token")
            headers["Authorization"] = f"Bearer {token}"

        for key, value in kwargs.items():
            if value is not None:
                headers[str(key)] = str(value)

        return headers

    def make_request_to_endpoint(
        self, endpoint, path, data=None, headers=None, is_download=False, files=None, **kwargs
    ):
        if headers is None:
            headers = {}
        if data is None:
            data = {}

        self.check_path_validity(path)

        headers = self.get_headers(path, files is not None, **headers)
        params = {str(key): _convert_param_value(value) for key, value in kwargs.items() if value is not None}

        self.rate_limit.wait()
        response = getattr(requests, path.method)(
            endpoint + path.path, headers=headers, data=data, params=params, files=files
        )
        try:
            response.raise_for_status()
        except Exception as e:
            try:
                err = response.json()["error"]
            except:
                err = None

            if err:
                raise RequestException(err) from e

            raise e

        if len(response.content) == 0:
            return

        return response.json() if not is_download else response

    def make_request(self, path, *args, **kwargs):
        return self.make_request_to_endpoint(self.base_url, path, *args, **kwargs)

    def get_auth_token(self, data):
        self.rate_limit.wait()
        return requests.post(self.token_url, data=data)

    @classmethod
    def from_async(cls, http: "AsynchronousHTTPHandler", auth: Optional["BaseAuthHandler"] = None):
        new_http = cls(auth, http.rate_limit.wait_time, http.rate_limit.limit, http.api_version)
        new_http.rate_limit._requests_sent = http.rate_limit._requests_sent
        new_http.base_url = http.base_url
        new_http.auth_url = http.auth_url
        new_http.token_url = http.token_url
        return new_http

    def as_async(self, auth: Optional["BaseAsynchronousAuthHandler"]) -> "AsynchronousHTTPHandler":
        from .asyncio.http import AsynchronousHTTPHandler

        return AsynchronousHTTPHandler.from_sync(self, auth)


class RateLimitHandler:
    __slots__ = (
        "wait_time",
        "limit",
        "_lock",
        "_waiting_lock",
        "_requests_sent",
    )

    def __init__(self, request_wait_time: float, limit_per_minute: int):
        self.wait_time: float = request_wait_time
        self.limit: int = limit_per_minute
        # for accessing non-thread-safe variables
        # intended to be used for short durations
        self._lock: threading.Lock = threading.Lock()
        # to make sure only one request is waiting at a time
        # okay to acquire for long periods of time
        self._waiting_lock: threading.Lock = threading.Lock()
        self._requests_sent: List[float] = []

    def wait(self):
        self._lock.acquire()

        if self.wait_time > 0:
            self._wait_with_wait_time()
        else:
            self._wait_without_wait_time()

        self._get_requests_sent().append(time.monotonic())

        self._lock.release()

    def _wait_with_wait_time(self):
        # acquiring _waiting_lock could take a bit
        # so let's release this one
        self._lock.release()
        # once acquired, we can choose the appropriate way to wait
        # without worry about race conditions. waiting one at a time
        # is okay since wait time between requests is > 0
        self._waiting_lock.acquire()

        self._lock.acquire()
        if len(requests_sent := self._get_requests_sent()) > 0:
            wait_time = max(0.0, self.wait_time - (time.monotonic() - requests_sent[-1]))
            if wait_time > 0:
                self._lock.release()
                time.sleep(wait_time)
                self._lock.acquire()

        self._waiting_lock.release()

    def _wait_without_wait_time(self):
        # under rate limit still, good to send
        if len(self._get_requests_sent()) < self.limit:
            return

        # acquiring _waiting_lock could take a bit
        # so let's release this one
        self._lock.release()
        self._waiting_lock.acquire()
        self._lock.acquire()

        # check again, then wait till oldest request expires past 1 minute
        if len(requests_sent := self._get_requests_sent()) >= self.limit:
            wait_time = max(0.0, 60.0 - (time.monotonic() - requests_sent[0]))
            if wait_time > 0:
                self._lock.release()
                time.sleep(wait_time)
                self._lock.acquire()

        self._waiting_lock.release()

    def _get_requests_sent(self):
        """expects self._lock is acquired when calling this function"""
        # update list
        while len(self._requests_sent) > 0 and time.monotonic() - self._requests_sent[0] >= 60:
            self._requests_sent.pop(0)

        return self._requests_sent



from .http import BaseHTTPHandler
from .objects import *
from .path import Path
from .enums import *
from .auth import BaseAuthHandler, AuthHandler, NoAuth
from .util import (
    parse_mods_arg,
    parse_enum_args,
    BeatmapsetSearchFilter,
    create_multipart_formdata,
    get_optional_list,
    get_optional,
)
from .results import *
from .scope import Scope

from typing import Union, Optional, Sequence, Dict, List
from datetime import datetime

try:
    import osrparse

    has_osrparse = True
except ImportError:
    has_osrparse = False


__all__ = ("Client",)


class Client:
    """
    Main object for interacting with osu!api, which uses synchronous requests.
    If you're looking for asynchronous requests, use :class:`AsynchronousClient`.

    .. WARNING::
        This class is not thread-safe. If you wish to use with threading,
        use a threading lock when making requests.

    .. WARNING::
        Regarding the ``limit_per_minute`` attribute below: do not change it unless you know what you are doing.
        The `terms of use <https://osu.ppy.sh/docs/#terms-of-use>`_ specify to avoid going over 60 requests per minute.

    :param auth:
        Typically will be an :class:`osu.auth.AuthHandler` or :class:`osu.auth.AsynchronousAuthHandler` object.
    :type auth: Optional[:class:`osu.auth.BaseAuthHandler`]
    :param request_wait_time: (Default 1.0)
        This defines the amount of time that the client should wait before making another request.
        It can make it easier to stay within the rate limits without using all your requests up quickly
        and then waiting forever to make another. It's most applicable in bot-type apps.
    :type request_wait_time: float
    :param limit_per_minute: (Default 60)
        This sets a cap on the number of requests the client is allowed to make within 1 minute of time.
        Don't change this before reading the above warning related to this attribute.
    :type limit_per_minute: int
    :param api_version:
        This parameter is here purely to expose the option.
        You likely don't need to mess with this.
        Format is 'yyyymmdd' (e.g. 20231030).
    :type api_version: Optional[str]
    """

    __slots__ = ("auth",)

    def __init__(
        self,
        auth: Optional[BaseAuthHandler] = None,
        request_wait_time: float = 1.0,
        limit_per_minute: int = 60,
        api_version: Optional[str] = None,
    ):
        self.auth: BaseAuthHandler = NoAuth() if auth is None else auth

        self.http.set_ratelimit(request_wait_time, limit_per_minute)
        self.http.set_api_version(api_version)

    @property
    def http(self) -> BaseHTTPHandler:
        return self.auth.http

    @classmethod
    def from_credentials(
        cls,
        client_id: int,
        client_secret: str,
        redirect_url: Optional[str],
        scope: Optional[Scope] = None,
        code: Optional[str] = None,
        request_wait_time: float = 1.0,
        limit_per_minute: int = 60,
        lazily_authenticate: bool = True,
    ) -> "Client":
        """
        Creates a :class:`Client` object from client id, client secret, redirect uri, and scope.

        **Parameters**

        client_id: int
            API client ID

        client_secret: str
            API client secret

        redirect_url: Optional[str]
            API redirect url

        scope: Optional[:class:`Scope`]
            Scopes to authenticate under. Default is :func:`Scope.default`, which is just the ``public`` scope.

        request_wait_time: float
            Read :class:`Client` for details.

        limit_per_minute: int
            Do not change default before reading details in :class:`Client`.

        lazily_authenticate: bool
            If true, the :class:`AuthHandler` won't authenticate with the api until
            a request is made that requires it.

        **Returns**

        :class:`Client`
        """
        if scope is None:
            scope = Scope.default()

        auth = AuthHandler(client_id, client_secret, redirect_url, scope)
        if not lazily_authenticate:
            auth.get_auth_token(code)
        return cls(auth, request_wait_time, limit_per_minute)

    @classmethod
    def from_client_credentials(cls, *args, **kwargs):
        """
        .. deprecated:: 2.2.0
            Use :func:`from_credentials`
        """
        return cls.from_credentials(*args, **kwargs)

    def set_api_version(self, version: str) -> None:
        """
        Sets x-api-version header to use when sending requests to the api.
        You shouldn't have to change it from the default, but if you need to
        then the function is available.

        **Parameters**

        version: str
            x-api-version header value in the format yyyymmdd
        """
        self.http.api_version = version

    def set_domain(self, domain: str) -> None:
        """
        Set domain to use for requests, such as "dev.ppy.sh"

        **Parameters**

        domain: str
        """
        self.http.set_domain(domain)

    def lookup_beatmap(
        self,
        checksum: Optional[str] = None,
        filename: Optional[str] = None,
        id: Optional[int] = None,
    ) -> Beatmap:
        """
        Returns beatmap.

        Requires OAuth and scope public

        **Parameters**

        checksum: Optional[str]
            A beatmap checksum.

        filename: Optional[str]
            A filename to lookup

        id: Optional[int]
            A beatmap ID to lookup

        **Returns**

        :class:`Beatmap`
        """
        return Beatmap(self.http.make_request(Path.beatmap_lookup(), checksum=checksum, filename=filename, id=id))

    def get_user_beatmap_score(
        self,
        beatmap: int,
        user: int,
        mode: Optional[Union[str, GameModeStr]] = None,
        mods: Optional[Sequence[str]] = None,
    ) -> BeatmapUserScore:
        """
        Returns a user's score on a Beatmap

        Requires OAuth and scope public

        **Parameters**

        beatmap: int
            Id of the beatmap

        user: int
            Id of the user

        mode: Optional[Union[str, :class:`GameModeStr`]]
            The game mode to get scores for

        mods: Optional[Sequence[str]]
            An array of matching mods, or none. Currently doesn't do anything.

        **Returns**

        :class:`BeatmapUserScore`
        """
        mode = parse_enum_args(mode)
        return BeatmapUserScore(
            self.http.make_request(Path.user_beatmap_score(beatmap, user), mode=mode, mods=mods), self.http.api_version
        )

    def get_user_beatmap_scores(
        self, beatmap: int, user: int, mode: Optional[Union[str, GameModeStr]] = None
    ) -> List[Union[LegacyScore, SoloScore]]:
        """
        Returns a user's scores on a Beatmap

        Requires OAuth and scope public

        **Parameters**

        beatmap: int
            Id of the beatmap

        user: int
            Id of the user

        mode: Optional[Union[str, :class:`GameModeStr`]]
            The game mode to get scores for

        **Returns**

        List[Union[:class:`LegacyScore`, :class:`SoloScore`]]
            Will be SoloScore unless using an older api version
        """
        mode = parse_enum_args(mode)
        resp = self.http.make_request(Path.user_beatmap_scores(beatmap, user), mode=mode)
        return [get_score_object(score, self.http.api_version) for score in resp["scores"]]

    def _parse_mods_list(self, mods) -> Optional[List[str]]:
        if mods is None:
            return
        return list(
            map(
                lambda mod: (
                    (Mod[mod.name].value if not isinstance(mod, Mod) else mod.value) if type(mod) != str else mod
                ),
                mods,
            )
        )

    def get_beatmap_scores(
        self,
        beatmap: int,
        mode: Optional[Union[str, GameModeStr]] = None,
        mods: Optional[Union[Mods, Sequence[Union[Mods, Mod, str]]]] = None,
        ranking_type: Optional[str] = None,
        legacy_only: Optional[bool] = None,
    ) -> BeatmapScores:
        """
        Returns the top scores for a beatmap

        Requires OAuth and scope public

        **Parameters**

        beatmap: int
            Id of the beatmap

        mode: Optional[Union[str, :class:`GameModeStr`]]
            The game mode to get scores for

        mods: Optional[Union[:class:`Mods`, Sequence[Union[:class:`Mods`, :class:`Mod`, str]]]]
            Must pass one of:
            a :class:`Mods` object,
            a list of string mod abbreviations,
            a list of :class:`Mods` objects,
            a list of :classL`Mod` objects

        ranking_type: Optional[str]
            Beatmap score ranking type. Currently doesn't do anything.

        legacy_only: Optional[bool]
            Whether or not to exclude lazer scores. Defaults to false.

        **Returns**

        :class:`BeatmapScores`
            :class:`osu.objects.LegacyScore` object inside includes ``user`` and the
            included user includes ``country`` and ``cover``.
        """
        mode = parse_enum_args(mode)
        mods = self._parse_mods_list(mods)
        return BeatmapScores(
            self.http.make_request(
                Path.beatmap_scores(beatmap),
                mode=mode,
                **{"mods[]": mods},
                type=ranking_type,
                legacy_only=1 if legacy_only else 0,
            ),
            self.http.api_version,
        )

    def get_beatmap(self, beatmap: int) -> Beatmap:
        """
        Gets beatmap data for the specified beatmap ID.

        Requires OAuth and scope public

        **Parameters**

        beatmap: int
            The ID of the beatmap

        **Returns**

        :class:`Beatmap`
            Includes attributes `beatmapset`, `beatmapset.ratings`, `failtimes`, `max_combo`.
        """
        return Beatmap(self.http.make_request(Path.beatmap(beatmap)))

    def get_beatmaps(self, ids: Optional[Sequence[int]] = None) -> List[Beatmap]:
        """
        Returns list of beatmaps.

        Requires OAuth and scope public

        **Parameters**

        ids: Optional[List[int]]
            Beatmap id to be returned. Specify once for each beatmap id requested.
            Up to 50 beatmaps can be requested at once.

        **Returns**

        List[:class:`Beatmap`]
            Includes attributes `beatmapset`, `beatmapset.ratings`, `failtimes`, `max_combo`.
        """
        results = self.http.make_request(Path.beatmaps(), **{"ids[]": list(ids)})
        return list(map(Beatmap, results["beatmaps"])) if results else []

    def get_beatmap_attributes(
        self,
        beatmap: int,
        mods: Optional[Union[int, Mods, Sequence[Union[str, Mods, int]]]] = None,
        ruleset: Optional[Union[str, GameModeStr]] = None,
        ruleset_id: Optional[Union[int, GameModeInt]] = None,
    ) -> BeatmapDifficultyAttributes:
        """
        Returns difficulty attributes of beatmap with specific mode and mods combination.

        Requires OAuth and scope public

        **Parameters**

        beatmap: int
            Beatmap id.

        mods: Optional[Union[int, Sequence[Union[str, :class:`Mods`, int]], :class:`Mods`]]
            Mod combination. Can be either a bitset of mods, a Mods enum, or array of any. Defaults to no mods.
            Some mods may cause the api to throw an HTTP 422 error depending on the map's gamemode.

        ruleset: Optional[Union[:class:`GameModeStr`, str]]
            Ruleset of the difficulty attributes. Only valid if it's the beatmap ruleset or the beatmap can be
            converted to the specified ruleset. Defaults to ruleset of the specified beatmap.

        ruleset_id: Optional[Union[:class:`GameModeInt`, int]]
            The same as `ruleset` but in integer form.

        **Returns**

        :class:`BeatmapDifficultyAttributes`
        """
        ruleset, ruleset_id = parse_enum_args(ruleset, ruleset_id)
        if ruleset is not None:
            mode = GameModeStr(ruleset)
        elif ruleset_id is not None:
            mode = GameModeInt(ruleset_id).get_str_equivalent()
        else:
            mode = GameModeStr.STANDARD
        return BeatmapDifficultyAttributes(
            self.http.make_request(
                Path.get_beatmap_attributes(beatmap),
                mods=parse_mods_arg(mods),
                ruleset=ruleset,
                ruleset_id=ruleset_id,
            ),
            mode,
        )

    def get_beatmapset(self, beatmapset_id: int) -> Beatmapset:
        """
        Get beatmapset by id.

        Requires OAuth and scope public

        **Parameters**

        beatmapset_id: int

        **Returns**

        :class:`Beatmapset`
        """
        return Beatmapset(self.http.make_request(Path.get_beatmapset(beatmapset_id)))

    def get_beatmapset_discussion_posts(
        self,
        beatmapset_discussion_id: Optional[int] = None,
        limit: Optional[int] = None,
        page: Optional[int] = None,
        sort: Optional[str] = None,
        types: Optional[Sequence[str]] = None,
        user: Optional[int] = None,
        with_deleted: Optional[str] = None,
        cursor: Optional[Dict[str, int]] = None,
    ) -> BeatmapsetDiscussionPostsResult:
        """
        Returns the posts of the beatmapset discussions

        Requires OAuth and scope public

        **Parameters**

        beatmapset_discussion_id: Optional[int]
            id of the BeatmapsetDiscussion

        limit: Optional[int]
            Maximum number of results

        page: Optional[int]
            Search results page.

        sort: Optional[str]
            `id_desc` for newest first; `id_asc` for oldest first. Defaults to `id_desc`

        type: Optional[Sequence[str]]
            `first`, `reply`, `system` are the valid values. Defaults to `reply`.

        user: Optional[int]
            The id of the user

        with_deleted: Optional[str]
            This param has no effect as api calls do not currently receive group permissions.

        cursor: Optional[Dict[str, int]]
            A cursor object received from a previous call to get_beatmapset_discussion_posts
            (:class:`BeatmapsetDiscussionPostsResult`.cursor)

        **Returns**

        :class:`BeatmapsetDiscussionsPostsResult`
        """
        if cursor is None:
            cursor = {}
        if "page" in cursor:
            page = cursor["page"]
        if "limit" in cursor:
            limit = cursor["limit"]
        resp = self.http.make_request(
            Path.beatmapset_discussion_posts(),
            beatmapset_discussion_id=beatmapset_discussion_id,
            limit=limit,
            page=page,
            sort=sort,
            user=user,
            with_deleted=with_deleted,
            **{"types[]": types},
        )
        return BeatmapsetDiscussionPostsResult(
            list(map(BeatmapsetCompact, resp["beatmapsets"])),
            list(map(BeatmapsetDiscussionPost, resp["posts"])),
            list(map(UserCompact, resp["users"])),
            resp["cursor_string"],
        )

    def get_beatmapset_discussion_votes(
        self,
        beatmapset_discussion_id: Optional[int] = None,
        limit: Optional[int] = None,
        page: Optional[int] = None,
        receiver: Optional[int] = None,
        score: Optional[int] = None,
        sort: Optional[str] = None,
        user: Optional[int] = None,
        with_deleted: Optional[str] = None,
        cursor: Optional[Dict[str, int]] = None,
    ) -> BeatmapsetDiscussionVotesResult:
        """
        Returns the votes given to beatmapset discussions

        Requires OAuth and scope public

        **Parameters**

        beatmapset_discussion_id: Optional[int]
            id of the BeatmapsetDiscussion

        limit: Optional[int]
            Maximum number of results

        page: Optional[int]
            Search results page.

        receiver: Optional[int]
            The id of the User receiving the votes.

        score: Optional[int]
            1 for upvote, -1 for downvote

        sort: Optional[str]
            `id_desc` for newest first; `id_asc` for oldest first. Defaults to `id_desc`

        user: Optional[int]
            The id of the User giving the votes.

        with_deleted: Optional[str]
            This param has no effect as api calls do not currently receive group permissions

        cursor: Optional[Dict[str, int]]
            A cursor object received from a previous call to get_beatmapset_discussion_votes
            (:class:`BeatmapsetDiscussionVotesResult`.cursor)

        **Returns**

        :class:`BeatmapsetDiscussionVotesResult`
        """
        if cursor is None:
            cursor = {}
        if "page" in cursor:
            page = cursor["page"]
        if "limit" in cursor:
            limit = cursor["limit"]
        resp = self.http.make_request(
            Path.beatmapset_discussion_votes(),
            beatmapset_discussion_id=beatmapset_discussion_id,
            limit=limit,
            receiver=receiver,
            score=score,
            page=page,
            sort=sort,
            user=user,
            with_deleted=with_deleted,
        )
        return BeatmapsetDiscussionVotesResult(
            list(map(BeatmapsetDiscussion, resp["discussions"])),
            list(map(BeatmapsetDiscussionVote, resp["votes"])),
            list(map(UserCompact, resp["users"])),
            resp["cursor"],
        )

    def get_beatmapset_discussions(
        self,
        beatmap_id: Optional[int] = None,
        beatmapset_id: Optional[int] = None,
        beatmapset_status: Optional[str] = None,
        limit: Optional[int] = None,
        message_types: Optional[Sequence[Union[str, MessageType]]] = None,
        only_unresolved: Optional[bool] = None,
        page: Optional[int] = None,
        sort: Optional[str] = None,
        user: Optional[int] = None,
        with_deleted: Optional[str] = None,
        cursor: Optional[Dict[str, int]] = None,
    ) -> BeatmapsetDiscussionsResult:
        """
        Returns a list of beatmapset discussions

        Requires OAuth and scope public

        **Parameters**

        beatmap_id: Optional[int]
            id of the beatmap

        beatmapset_id: Optional[int]
            id of the beatmapset

        beatmapset_status: Optional[str]
            One of `all`, `ranked`, `qualified`, `disqualified`, `never_qualified`. Defaults to `all`.

        limit: Optional[int]
            Maximum number of results.

        message_types: Optional[Sequence[Union[str, :class:`MessageType`]]]
            None defaults to all types.

        only_unresolved: Optional[bool]
            true to show only unresolved issues; false, otherwise. Defaults to false.

        page: Optional[int]
            Search result page.

        sort: Optional[str]
            `id_desc` for newest first; `id_asc` for oldest first. Defaults to `id_desc`.

        user: Optional[int]
            The id of the User.

        with_deleted: Optional[str]
            This param has no effect as api calls do not currently receive group permissions.

        cursor: Optional[Dict[str, int]]
            A cursor object received from a previous call to get_beatmapset_discussions
            (:class:`BeatmapsetDiscussionsResult`.cursor)

        **Returns**

        :class:`BeatmapsetDiscussionsResult`
        """
        if cursor is None:
            cursor = {}
        if "page" in cursor:
            page = cursor["page"]
        if "limit" in cursor:
            limit = cursor["limit"]
        params = {}
        if message_types is not None:
            message_types = list(map(lambda t: t.value if isinstance(t, MessageType) else t, message_types))
            params = {"message_types[]": message_types}
        resp = self.http.make_request(
            Path.beatmapset_discussions(),
            beatmap_id=beatmap_id,
            beatmapset_id=beatmapset_id,
            beatmapset_status=beatmapset_status,
            limit=limit,
            only_unresolved=only_unresolved,
            page=page,
            sort=sort,
            user=user,
            with_deleted=with_deleted,
            **params,
        )
        return BeatmapsetDiscussionsResult(
            list(map(Beatmap, resp["beatmaps"])),
            list(map(BeatmapsetDiscussion, resp["discussions"])),
            list(map(BeatmapsetDiscussion, resp["included_discussions"])),
            list(map(UserCompact, resp["users"])),
            ReviewsConfig(resp["reviews_config"]),
            resp["cursor"],
        )

    def get_changelog_build(self, stream: str, build: str) -> Build:
        """
        Returns details of the specified build.

        **Parameters**

        stream: str
            Update stream name.

        build: str
            Build version.

        **Returns**

        A :class:`Build` with `changelog_entries`, `changelog_entries.github_user`, and `versions` included.
        """
        return Build(self.http.make_request(Path.get_changelog_build(stream, build)))

    def get_changelog_listing(
        self,
        start: Optional[str] = None,
        max_id: Optional[int] = None,
        stream: Optional[str] = None,
        end: Optional[str] = None,
        message_formats: Optional[Sequence[str]] = None,
    ) -> ChangelogListingResult:
        """
        Returns a listing of update streams, builds, and changelog entries.

        **Parameters**

        start: Optional[str]
            Minimum build version.

        max_id: Optional[int]
            Maximum build ID.

        stream: Optional[str]
            Stream name to return builds from.

        end: Optional[str]
            Maximum build version.

        message_formats: Optional[Sequence[str]]
            `html`, `markdown`. Default to both.

        **Returns**

        :class:`ChangelogListingResult`
        """
        response = self.http.make_request(
            Path.get_changelog_listing(),
            max_id=max_id,
            stream=stream,
            to=end,
            **{"from": start, "message_formats[]": message_formats},
        )
        return ChangelogListingResult(
            list(map(Build, response["builds"])),
            list(map(UpdateStream, response["streams"])),
            ChangelogListingSearch(
                response["search"]["from"],
                response["search"]["to"],
                response["search"]["limit"],
                response["search"]["max_id"],
                response["search"]["stream"],
            ),
        )

    def lookup_changelog_build(
        self,
        changelog: str,
        key: Optional[str] = None,
        message_formats: Optional[Sequence[str]] = None,
    ) -> Build:
        """
        Returns details of the specified build.

        **Parameters**

        changelog: str
            Build version, update stream name, or build ID.

        key: Optional[str]
            Leave blank to query by build version or stream name, or `id` to query by build ID.

        message_formats: Optional[Sequence[str]]
            `html`, `markdown`. Default to both.

        **Returns**

        A :class:`Build` with changelog_entries, changelog_entries.github_user, and versions included.
        """
        return Build(
            self.http.make_request(
                Path.lookup_changelog_build(changelog), key=key, **{"message_formats[]": message_formats}
            )
        )

    def get_comments(
        self,
        commentable_type: Optional[Union[ObjectType, str]] = None,
        commentable_id: Optional[int] = None,
        cursor: Optional[dict] = None,
        parent_id: Optional[int] = None,
        sort: Optional[Union[str, CommentSort]] = None,
    ) -> CommentBundle:
        """
        Returns a list comments and their replies up to 2 levels deep.

        Does not require OAuth

        **Parameter**

        commentable_type: Optional[Union[:class:`ObjectType`, str]
            The type of resource to get comments for. Must be of the following types:
            beatmapset, build, news_post

        commentable_id: Optional[int]
            The id of the resource to get comments for. Id correlates with commentable_type.

        cursor: Optional[:class:`dict`]
            Pagination option. Returned in :class:`CommentSort` for next request.

        parent_id: Optional[int]
            Limit to comments which are reply to the specified id. Specify 0 to get top level comments.

        sort: Optional[Union[str, :class:`CommentSort`]]
            Sort option as defined in :class:`CommentSort`.
            Defaults to new for guests and user-specified default when authenticated.

        **Returns**

        :class:`CommentBundle`
            pinned_comments is only included when commentable_type and commentable_id are specified.
        """
        commentable_type, sort = parse_enum_args(commentable_type, sort)
        return CommentBundle(
            self.http.make_request(
                Path.get_comments(),
                commentable_type=commentable_type,
                commentable_id=commentable_id,
                parent_id=parent_id,
                sort=sort,
                **(cursor if cursor else {}),
            )
        )

    def get_comment(self, comment: int) -> CommentBundle:
        """
        Gets a comment and its replies up to 2 levels deep.

        Does not require OAuth

        **Parameters**

        comment: int
            Comment id

        **Returns**

        :class:`CommentBundle`
        """
        return CommentBundle(self.http.make_request(Path.get_comment(comment)))

    def reply_topic(self, topic: int, body: str) -> ForumPost:
        """
        Create a post replying to the specified topic.

        Requires OAuth, scope forum.write, and a user (authorization code grant or delegate scope)

        **Parameters**

        topic: int
            Id of the topic to be replied to.

        body: str
            Content of the reply post.

        **Returns**

        :class:`ForumPost`
            body attributes included
        """
        data = {"body": body}
        return ForumPost(self.http.make_request(Path.reply_topic(topic), files=create_multipart_formdata(data)))

    def create_topic(
        self,
        body: str,
        forum_id: int,
        title: str,
        with_poll: Optional[bool] = None,
        hide_results: Optional[bool] = None,
        length_days: Optional[int] = None,
        max_options: Optional[int] = None,
        poll_options: Optional[List[str]] = None,
        poll_title: Optional[str] = None,
        vote_change: Optional[bool] = None,
    ) -> CreateTopicResult:
        """
        Create a new topic.

        Requires OAuth, scope forum.write, and a user (authorization code grant or delegate scope)

        **Parameters**

        body: str
            Content of the topic.

        forum_id: int
            Forum to create the topic in.

        title: str
            Title of the topic.

        with_poll: Optional[bool]
            Enable this to also create poll in the topic (default: false).

        hide_results: Optional[bool]
            Enable this to hide result until voting period ends (default: false).

        length_days: Optional[int]
            Number of days for voting period. 0 means the voting will never ends (default: 0).
            This parameter is required if hide_results option is enabled.

        max_options: Optional[int]
            Maximum number of votes each user can cast (default: 1).

        poll_options: Optional[List[str]]
            List of voting options. BBCode is supported.

        poll_title: Optional[str]
            Title of the poll.

        vote_change: Optional[bool]
            Enable this to allow user to change their votes (default: false).

        **Returns**

        :class:`CreateTopicResult`
        """
        data = {
            "body": body,
            "forum_id": forum_id,
            "title": title,
            "with_poll": with_poll,
        }
        if with_poll:
            if poll_options is None or poll_title is None:
                raise TypeError("poll_options and poll_title are required since the topic has a poll.")
            for k, v in {
                "forum_topic_poll[hide_results]": hide_results,
                "forum_topic_poll[length_days]": length_days,
                "forum_topic_poll[max_options]": max_options,
                "forum_topic_poll[poll_options]": "\n".join(poll_options) if poll_options is not None else None,
                "forum_topic_poll[poll_title]": poll_title,
                "forum_topic_poll[vote_change]": vote_change,
            }.items():
                if v is not None:
                    data[k] = v

        resp = self.http.make_request(Path.create_topic(), files=create_multipart_formdata(data))
        return CreateTopicResult(ForumTopic(resp["topic"]), ForumPost(resp["post"]))

    def get_topic_and_posts(
        self,
        topic: int,
        cursor: Optional[str] = None,
        sort: Optional[str] = None,
        limit: Optional[int] = None,
        start: Optional[int] = None,
        end: Optional[int] = None,
    ) -> GetTopicAndPostsResult:
        """
        Get topic and its posts.

        Requires OAuth and scope public

        **Parameters**

        topic: int
            Id of the topic.

        cursor: Optional[str]
            Parameter for pagination.

        sort: Optional[str]
            Post sorting option. Valid values are id_asc (default) and id_desc.

        limit: Optional[int]
            Maximum number of posts to be returned (20 default, 50 at most).

        start: Optional[int]
            First post id to be returned with sort set to id_asc.
            This parameter is ignored if cursor_string is specified.

        end: Optional[int]
            First post id to be returned with sort set to id_desc.
            This parameter is ignored if cursor_string is specified.

        **Returns**

        :class:`GetTopicAndPostsResult`
        """
        resp = self.http.make_request(
            Path.get_topic_and_posts(topic),
            **(cursor if cursor else {}),
            sort=sort,
            limit=limit,
            start=start,
            end=end,
        )
        return GetTopicAndPostsResult(
            resp["cursor_string"],
            resp["search"],
            ForumTopic(resp["topic"]),
            list(map(ForumPost, resp["posts"])),
        )

    def edit_topic(self, topic: int, topic_title: str) -> ForumTopic:
        """
        Edit topic. Only title can be edited through this endpoint.

        Requires OAuth and scope forum.write

        **Parameters**

        topic: int
            Id of the topic.

        topic_title: str
            New topic title.

        **Returns**

        :class:`ForumTopic`
        """
        data = {"forum_topic[topic_title]": topic_title}
        return ForumTopic(self.http.make_request(Path.edit_topic(topic), files=create_multipart_formdata(data)))

    def edit_post(self, post: int, body: str) -> ForumPost:
        """
        Edit specified forum post.

        Requires OAuth and scope forum.write

        post: int
            Id of the post.

        body: str
            New post content in BBCode format.

        **Returns**

        :class:`ForumPost`
        """
        data = {"body": body}
        return ForumPost(self.http.make_request(Path.edit_post(post), files=create_multipart_formdata(data)))

    def search(
        self,
        mode: Optional[Union[str, WikiSearchMode]] = None,
        query: Optional[str] = None,
        page: Optional[int] = None,
    ) -> SearchResult:
        """
        Searches users and wiki pages.

        Requires OAuth and scope public

        **Parameters**

        mode: Optional[Union[str, :class:`WikiSearchMode`]]
            Either all, user, or wiki_page. Default is all.

        query: Optional[str]
            Search keyword.

        page: Optional[int]
            Search result page. Ignored for mode all.

        **Returns**

        :class:`SearchResult`
        """
        mode = parse_enum_args(mode)
        resp = self.http.make_request(Path.search(), mode=mode, query=query, page=page)
        return SearchResult(
            get_optional_list(resp.get("user", {}), "data", UserCompact),
            get_optional_list(resp.get("wiki_page", {}), "data", WikiPage),
        )

    def get_scores(
        self,
        room: int,
        playlist: int,
        limit: Optional[int] = None,
        sort: Optional[Union[str, MultiplayerScoresSort]] = None,
        cursor: Optional[str] = None,
    ) -> MultiplayerScores:
        """
        If you're looking for the endpoint to get a list of 1000 recent scores,
        check :func:`get_all_scores`.

        Requires OAuth, scope public, and a user (authorization code grant or delegate scope)

        **Parameters**

        room: int
            Id of the room.

        playlist: int
            Id of the playlist item.

        limit: Optional[int]
            Number of scores to be returned.

        sort: Optional[Union[str, :class:`MultiplayerScoresSort`]]

        cursor: Optional[str]
            :class:`MultiplayerScores`.cursor value from a previous call to get next page.

        **Returns**

        :class:`MultiplayerScores`
        """
        sort = parse_enum_args(sort)
        return MultiplayerScores(
            self.http.make_request(
                Path.get_scores(room, playlist),
                limit=limit,
                sort=sort,
                cursor_string=cursor,
            )
        )

    def get_news_listing(
        self,
        limit: Optional[int] = None,
        year: Optional[int] = None,
        cursor: Optional[dict] = None,
    ) -> GetNewsListingResult:
        """
        Returns a list of news posts and related metadata.

        **Parameters**

        limit: Optional[int]
            Maximum number of posts (12 default, 1 minimum, 21 maximum).

        year: Optional[int]
            Year to return posts from.

        cursor: Optional[str]
            Cursor for pagination.

        **Returns**

        :class:`GetNewsListingResult`
        """
        resp = self.http.make_request(Path.get_news_listing(), limit=limit, year=year, cursor=cursor)
        return GetNewsListingResult(
            resp["cursor_string"],
            list(map(NewsPost, resp["news_posts"])),
            NewsSidebar(
                resp["news_sidebar"]["current_year"],
                list(map(NewsPost, resp["news_sidebar"]["news_posts"])),
                resp["news_sidebar"]["years"],
            ),
            SearchInfo(resp["search"]["sort"], resp["search"]["limit"], None, None),
        )

    def get_news_post(self, news: str, key: Optional[str] = None) -> NewsPost:
        """
        Returns details of the specified news post.

        **Parameters**

        news: class:`str`
            News post slug or ID.

        key: Optional[str]
            Unset to query by slug, or `id` to query by ID.

        **Returns**

        Returns a :class:`NewsPost` with content and navigation included.
        """
        return NewsPost(self.http.make_request(Path.get_news_post(news), key=key))

    def revoke_current_token(self) -> None:
        """
        Revokes currently authenticated token.

        Requires OAuth
        """
        self.http.make_request(Path.revoke_current_token())

    def get_ranking(
        self,
        mode: Union[str, GameModeStr],
        type: Union[str, RankingType],
        country: Optional[str] = None,
        cursor: Optional[dict] = None,
        filter: Optional[str] = None,
        spotlight: Optional[int] = None,
        variant: Optional[str] = None,
    ) -> Union[Rankings[UserStatistics], Rankings[CountryStatistics], Rankings[UserTeamStatistics], SpotlightRankings]:
        """
        Gets the current ranking for the specified type and game mode.

        Requires OAuth and scope public

        mode: Union[str, :class:`GameModeStr`]

        type: Union[str, :class:`RankingType`]
            :class:`RankingType`

        country: Optional[str]
            Filter ranking by country code. Only available for `type` of `performance`.

        cursor: Optional[:class:`dict`]

        filter: Optional[str]
            Either `all` (default) or `friends`.

        spotlight: Optional[int]
            The id of the spotlight if `type` is `charts`.
            Ranking for latest spotlight will be returned if not specified.

        variant: Optional[str]
            Filter ranking to specified mode variant.
            For `mode` of `mania`, it's either `4k` or `7k`. Only available for `type` of `performance`.

        **Returns**

        Union[:class:`Rankings`[:class:`UserStatistics`], :class:`Rankings`[:class:`CountryStatistics`]
        , :class:`Rankings`[:class:`UserTeamStatistics`], :class:`SpotlightRankings`]
            Ranking type that depends on `type` argument
        """
        mode, type = parse_enum_args(mode, type)
        data = self.http.make_request(
            Path.get_ranking(mode, type),
            country=country,
            **(cursor if cursor else {}),
            filter=filter,
            spotlight=spotlight,
            variant=variant,
        )
        return (
            SpotlightRankings(data)
            if type == "charts"
            else Rankings(data, {"team": UserTeamStatistics, "country": CountryStatistics}.get(type, UserStatistics))
        )

    def get_spotlights(self) -> Spotlights:
        """
        Gets the list of spotlights.

        Requires OAuth and scope public

        **Returns**

        :class:`Spotlights`
        """
        return Spotlights(self.http.make_request(Path.get_spotlights()))

    def get_own_data(self, mode: Union[str, GameModeStr] = "") -> User:
        """
        Similar to get_user but with authenticated user (token owner) as user id.

        Requires OAuth, scope identify, and a user (authorization code grant or delegate scope)

        **Parameters**

        mode: Optional[str, :class:`GameModeStr`]
            GameMode. User default mode will be used if not specified.

        **Returns**

        See return for get_user
        """
        mode = parse_enum_args(mode)
        return User(self.http.make_request(Path.get_own_data(mode)))

    def get_user_kudosu(self, user: int, limit: Optional[int] = None, offset: Optional[int] = None):
        """
        Returns kudosu history.

        Requires OAuth and scope public

        **Parameters**

        user: int
            Id of the user.

        limit: Optional[int]
            Maximum number of results.

        offset: Optional[int]
            Result offset for pagination.

        **Returns**

        Sequence[:class:`KudosuHistory`]
        """
        return list(
            map(
                KudosuHistory,
                self.http.make_request(Path.get_user_kudosu(user), limit=limit, offset=offset),
            )
        )

    def get_user_scores(
        self,
        user: int,
        type: Union[UserScoreType, str],
        include_fails: Optional[bool] = False,
        mode: Optional[Union[str, GameModeStr]] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
    ) -> List[Union[LegacyScore, SoloScore]]:
        """
        This endpoint returns the scores of specified user.

        Requires OAuth and scope public

        **Parameters**

        user: int
            Id of the user.

        type: Union[:class:`UserScoreType` str]
            Score type. Must be one of `best`, `firsts`, `recent`, `pinned`

        include_fails: Optional[bool]
            Only for recent scores, include scores of failed plays. Defaults to False.

        mode: Optional[Union[:class:`GameModeStr`, str]]
            game mode of the scores to be returned. Defaults to the specified user's mode.

        limit: Optional[int]
            Maximum number of results.

        offset: Optional[int]
            Result offset for pagination.

        **Returns**

        Sequence[Union[:class:`LegacyScore`, :class:`SoloScore`]]
            Includes attributes `beatmap`, `beatmapset`. Additionally includes `weight` if `type` is `best`.
        """
        mode, type = parse_enum_args(mode, type)
        return [
            get_score_object(score, self.http.api_version)
            for score in self.http.make_request(
                Path.get_user_scores(user, type),
                include_fails=int(include_fails),
                mode=mode,
                limit=limit,
                offset=offset,
            )
        ]

    def get_user_beatmaps(
        self,
        user: int,
        type: Union[str, UserBeatmapType],
        limit: Optional[int] = None,
        offset: Optional[int] = None,
    ) -> List[Union[BeatmapPlaycount, Beatmapset]]:
        """
        Returns the beatmaps of specified user.

        Requires OAuth and scope public

        **Parameters**

        user: int
            Id of the user.

        type: Union[str, :class:`UserBeatmapType`]
            Beatmap type. Can be one of `favourite`, `graveyard`, `guest`, `loved`, `most_played`, `nominated`, `pending`, `ranked`.

        limit: Optional[int]
            Maximum number of results.

        offset: Optional[int]
            Result offset for pagination.

        **Returns**

        List[Union[:class:`BeatmapPlaycount`, :class:`Beatmapset`]]
            :class:`BeatmapPlaycount` for `type` `most_played` and :class:`Beatmapset` for any other type.
        """
        type = parse_enum_args(type)
        return list(
            map(
                BeatmapPlaycount if type == "most_played" else Beatmapset,
                self.http.make_request(Path.get_user_beatmaps(user, type), limit=limit, offset=offset),
            )
        )

    def get_user_recent_activity(
        self, user: int, limit: Optional[int] = None, offset: Optional[int] = None
    ) -> List[EVENT_TYPE]:
        """
        Returns recent activity.

        Requires OAuth and scope public

        **Parameters**

        user: int
            Id of the user.

        limit: Optional[int]
            Maximum number of results.

        offset: Optional[int]
            Result offset for pagination.

        **Returns**

        List[:class:`Event`]
            list of :class:`Event` objects
        """
        return list(
            map(
                get_event_object,
                self.http.make_request(Path.get_user_recent_activity(user), limit=limit, offset=offset),
            )
        )

    def get_user(
        self,
        user: Union[int, str],
        mode: Optional[Union[str, GameModeStr]] = "",
        key: Optional[str] = None,
    ) -> User:
        """
        This endpoint returns the detail of specified user.

        NOTE: It's highly recommended to pass key parameter
        to avoid getting unexpected result (mainly when
        looking up user with numeric username or nonexistent user id).

        Requires OAuth and scope public

        **Parameters**

        user: Union[int, str]
            Id or username of the user. Id lookup is prioritised unless key parameter is specified.
            Previous usernames are also checked in some cases.

        mode: Optional[Union[str, :class:`GameModeStr`]
            User default mode will be used if not specified.

        key: Optional[str]
            **DEPRECATED**
            It's recommended to prefix usernames with @ instead of setting key

            Type of user passed in url parameter. Can be either `id` or `username`
            to limit lookup by their respective type. Passing empty or invalid
            value will result in id lookup followed by username lookup if not found.

        **Returns**

        :class:`User`
            Includes attributes `account_history`, `active_tournament_banner`, `badges`,
            `beatmap_playcounts_count`, `favourite_beatmapset_count`, `follower_count`,
            `graveyard_beatmapset_count`, `groups`, `loved_beatmapset_count`, `mapping_follower_count`,
            `monthly_playcounts`, `page`, `pending_beatmapset_count`, `previous_usernames`,
            `rank_highest`, `rank_history`, `ranked_beatmapset_count`, `replays_watched_counts`,
            `scores_best_count`, `scores_first_count`, `scores_recent_count`, `statistics`,
            `statistics.country_rank`, `statistics.rank`, `statistics.variants`, `support_level`,
            `user_achievements`.
        """
        mode = parse_enum_args(mode)
        user = f"@{user}" if key is not None and key.lower() == "username" else user
        return User(self.http.make_request(Path.get_user(user, mode)))

    def get_users(self, ids: Sequence[int], include_variant_statistics: Optional[bool] = None) -> List[UserCompact]:
        """
        Returns list of users.
        Can get maximum 50 at a time.

        Requires OAuth and scope public

        **Parameters**

        ids: Sequence[int]
            User id to be returned. Specify once for each user id requested.
            Up to 50 users can be requested at once.

        include_variant_statistics: Optional[bool]
            Whether to additionally include `statistics_rulesets.variants`. Defaults to false.

        **Returns**

        Sequence[:class:`UserCompact`]
            Includes attributes: country, cover, groups, statistics_rulesets.
        """
        res = self.http.make_request(
            Path.get_users(), **{"ids[]": ids}, include_variant_statistics=include_variant_statistics
        )
        return list(map(UserCompact, res["users"]))

    def lookup_users(self, users: List[Union[int, str]]):
        """
        Lookup users by a mix of user ids and usernames.
        Can lookup maximum 50 at a time.

        ids: Sequence[Union[int, str]]
            Can be a list of user ids and usernames.
            Usernames should be prefixed with "@" to make sure they're interpreted as usernames by the api.

        **Returns**

        Sequence[:class:`UserCompact`]
        """
        res = self.http.make_request(Path.lookup_users(), **{"ids[]": users})
        return list(map(UserCompact, res["users"]))

    def get_wiki_page(self, locale: str, path: str) -> WikiPage:
        """
        The wiki article or image data.

        No OAuth required.

        **Parameters**

        locale: str
            Two-letter language code of the wiki page.

        path: str
            The path name of the wiki page.

        **Returns**

        :class:`WikiPage`
        """
        return WikiPage(self.http.make_request(Path.get_wiki_page(locale, path)))

    def get_beatmapset_events(
        self,
        page: Optional[int] = None,
        limit: Optional[int] = None,
        sort: Optional[Union[str, BeatmapsetEventSort]] = None,
        type: Optional[Union[str, BeatmapsetEventType]] = None,
        min_date: Optional[Union[str, datetime]] = None,
        max_date: Optional[Union[str, datetime]] = None,
    ) -> GetBeatmapsetEventsResult:
        """
        Returns a list of beatmapset events.

        Requires OAuth and scope public.

        **Parameters**

        page: Optional[int]

        limit: Optional[int]

        sort: Optional[Union[str, :class:`BeatmapsetEventSort`]]
            Specified a sort order.

        type: Optional[Union[str, :class:`BeatmapsetEventType`]]
            Specifies for only a certain type of event to be returned.

        **Returns**

        :class:`GetBeatmapsetEventsResult`
        """
        sort, type = parse_enum_args(sort, type)
        if isinstance(min_date, datetime):
            min_date = min_date.isoformat()
        if isinstance(max_date, datetime):
            max_date = max_date.isoformat()
        resp = self.http.make_request(
            Path.get_beatmapset_events(),
            page=page,
            limit=limit,
            sort=sort,
            type=type,
            min_date=min_date,
            max_date=max_date,
        )
        return GetBeatmapsetEventsResult(
            list(map(BeatmapsetEvent, resp["events"])),
            Review(resp["reviewsConfig"]),
            list(map(UserCompact, resp["users"])),
        )

    def get_matches(
        self,
        limit: Optional[int] = None,
        sort: Optional[Union[str, MatchSort]] = None,
        cursor: Optional[Dict] = None,
        active: Optional[bool] = None,
    ) -> GetMatchesResult:
        """
        Returns a list of matches.

        Requires OAuth and scope public.

        **Parameters**

        limit: Optional[int]

        sort: Optional[Union[str, :class:`MatchSort`]]

        cursor: Optional[Dict]
            Dictionary containing one key: `match_id`.
            Can be obtained from a previous call to this function or manually created.

        active: Optional[bool]
            Specifies whether to return only active matches. Default returns active and unactive matches.

        **Returns**

        :class:`GetMatchesResult`
        """
        match_id = cursor.get("match_id") if cursor is not None else None
        sort = parse_enum_args(sort)
        resp = self.http.make_request(
            Path.get_matches(), limit=limit, sort=sort, active=active, **{"cursor[match_id]": match_id}
        )
        return GetMatchesResult(list(map(Match, resp["matches"])), resp["params"], resp["cursor"])

    def get_match(
        self, match_id: int, before: Optional[int] = None, after: Optional[int] = None, limit: Optional[int] = None
    ) -> MatchExtended:
        """
        Returns a match by id.

        Requires OAuth and scope public.

        **Parameters**

        match_id: int
            The match id.

        before: Optional[int]
            Filter for match events before the specified MatchEvent.id.

        after: Optional[int]
            Filter for match events after the specified MatchEvent.id.

        limit: Optional[int]
            Maximum number of match events to return. 100 default, 101 maximum.

        **Returns**

        :class:`Match`
        """
        return MatchExtended(
            self.http.make_request(Path.get_match(match_id), before=before, after=after, limit=limit),
            self.http.api_version,
        )

    def get_rooms(
        self,
        mode: Union[str, GameModeStr] = "",
        sort: Optional[Union[str, RoomSort]] = None,
        limit: Optional[int] = None,
        room_type: Optional[Union[RoomType, str]] = None,
        category: Optional[Union[RoomCategory, str]] = None,
        filter_mode: Optional[Union[RoomFilterMode, str]] = None,
    ) -> List[Room]:
        """
        Returns a list of rooms.

        Requires OAuth, scope public, and a user (authorization code grant or delegate scope).

        **Parameters**

        mode: Optional[Union[str, :class:`GameModeStr`]]
            Game mode to filter rooms by.

        sort: Optional[Union[str, :class:`RoomSort`]]
            Sort rooms by.

        limit: Optional[int]
            max number of rooms to return

        room_type: Optional[Union[:class:`RoomType`, str]]
            type of room to look for

        category: Optional[Union[:class:`RoomCategory`, str]]
            type of category of room to look for

        filter_mode: Optional[Union[:class:`RoomFilterMode`, str]]
        """
        mode, sort, room_type, category, filter_mode = parse_enum_args(mode, sort, room_type, category, filter_mode)
        return list(
            map(
                Room,
                self.http.make_request(
                    Path.get_rooms(mode),
                    sort=sort,
                    limit=limit,
                    type_group=room_type,
                    category=category,
                    mode=filter_mode,
                ),
            )
        )

    def get_seasonal_backgrounds(self) -> SeasonalBackgrounds:
        """
        Get the season backgrounds.

        Doesn't require OAuth

        **Returns**

        :class:`SeasonalBackgrounds`
        """
        return SeasonalBackgrounds(self.http.make_request(Path.get_seasonal_backgrounds()))

    def get_room(self, room_id: int) -> Room:
        """
        Returns a room by id.

        Requires OAuth and scope public.

        **Parameters**

        room_id: int
            The room id.

        **Returns**

        :class:`Room`
        """
        return Room(self.http.make_request(Path.get_room(room_id)))

    def get_score_by_id(self, mode: Union[str, GameModeStr], score_id) -> Union[LegacyScore, SoloScore]:
        """
        Returns a score by id.

        Requires OAuth and scope public.

        **Parameters**

        mode: Union[str, :class:`GameModeStr`]

        score_id: int

        **Returns**

        Union[:class:`SoloScore`, :class:`LegacyScore`]
            Should be a SoloScore, unless for some strange reason it's not
        """
        mode = parse_enum_args(mode)
        return get_score_object(self.http.make_request(Path.get_score_by_id(mode, score_id)), self.http.api_version)

    def get_score_by_id_only(self, score_id: int) -> Union[LegacyScore, SoloScore]:
        """
        Returns a score by id, not requiring a mode.

        Requires OAuth and scope public.

        **Parameters**

        score_id: int

        **Returns**

        Union[:class:`SoloScore`, :class:`LegacyScore`]
            Should be a SoloScore, unless for some strange reason it's not
        """
        return get_score_object(self.http.make_request(Path.get_score_by_id_only(score_id)), self.http.api_version)

    def search_beatmapsets(
        self, filters: Optional[BeatmapsetSearchFilter] = None, page: Optional[int] = None
    ) -> BeatmapsetSearchResult:
        """
        Search for beatmapsets.

        Requires OAuth and scope public.

        **Attributes**

        filters: Optional[:class:`BeatmapsetSearchFilter`]

        page: Optional[int]

        **Returns**

        :class:`BeatmapsetSearchResult`
        """
        if filters is None:
            filters = {}
        if isinstance(filters, BeatmapsetSearchFilter):
            filters = filters.filters
        resp = self.http.make_request(Path.beatmapset_search(), page=page, **filters)
        return BeatmapsetSearchResult(
            list(map(Beatmapset, resp["beatmapsets"])),
            resp["cursor"],
            resp["search"],
            resp["recommended_difficulty"],
            resp["error"],
            resp["total"],
        )

    def get_room_leaderboard(self, room_id: int) -> GetRoomLeaderboardResult:
        """
        Return a room's leaderboard. The :class:`UserScoreAggregate` objects returned under the "leaderboard"
        key contain the "user" attribute. The :class:`UserScoreAggregate` object under the "user_score" key
        contains the "user" and "position" attributes.

        Requires OAuth, scope public, and a user (authorization code grant or delegate scope).

        **Parameters**

        room_id: int

        **Returns**

        :class:`GetRoomLeaderboard`
        """
        resp = self.http.make_request(Path.get_room_leaderboard(room_id))
        return GetRoomLeaderboardResult(
            list(map(UserScoreAggregate, resp["leaderboard"])), get_optional(resp, "user_score", UserScoreAggregate)
        )

    def get_replay_data(
        self, mode: Optional[Union[GameModeStr, str]], score_id: int, use_osrparse: bool = True
    ) -> Union["osrparse.Replay", bytes]:
        """
        Returns replay data for a score. Use :func:`Client.get_replay_data_by_id_only` for only the id, or specify
        None to the mode attribute.

        Requires OAuth, scope public, and a user (authorization code grant or delegate scope).

        Requires osu.py is installed with the 'replay' feature if use_osrparse is true.

        **Parameters**

        mode: Optional[Union[str, :class:`GameModeStr`]]

        score_id: int

        use_osrparse: bool
            If true, returns an :class:`osrparse.Replay` object. Defaults to true.

        **Returns**

        Union[:class:`osrparse.Replay`, :class:`bytes`]
        """
        if mode is None:
            return self.get_replay_data_by_id_only(score_id, use_osrparse=use_osrparse)

        if not has_osrparse and use_osrparse:
            raise RuntimeError(
                "osrparse is required to call get_replay_data. "
                "Install osu.py with the 'replay' feature to use this function."
            )

        mode = parse_enum_args(mode)
        data = self.http.make_request(Path.get_replay_data(mode, score_id), is_download=True).content
        return osrparse.Replay.from_string(data) if use_osrparse else data

    def get_replay_data_by_id_only(self, score_id: int, use_osrparse: bool = True) -> Union["osrparse.Replay", bytes]:
        """
        Returns replay data for a score. Use :func:`Client.get_replay_data` for score ids that require
        specifying the game mode too.

        Requires OAuth, scope public, and a user (authorization code grant or delegate scope).

        Requires osu.py is installed with the 'replay' feature if use_osrparse is true.

        **Parameters**

        score_id: int

        use_osrparse: bool
            If true, returns an :class:`osrparse.Replay` object. Defaults to true.

        **Returns**

        Union[:class:`osrparse.Replay`, :class:`bytes`]
        """
        if not has_osrparse and use_osrparse:
            raise RuntimeError(
                "osrparse is required to call get_replay_data. "
                "Install osu.py with the 'replay' feature to use this function."
            )

        data = self.http.make_request(Path.get_replay_data_by_id_only(score_id), is_download=True).content
        return osrparse.Replay.from_string(data) if use_osrparse else data

    def get_friends(self) -> Union[List[UserRelation], List[UserCompact]]:
        """
        Returns a list of friends.

        Requires OAuth, scope friends.read, and a user (authorization code grant or delegate scope).

        **Returns**

        Union[List[:class:`UserCompact`], List[:class:`UserRelation`]]
            will be UserRelation objects if using default api version header
        """
        ret = self.http.make_request(Path.get_friends())
        if len(ret) == 0:
            return []

        return list(map(UserRelation if "target_id" in ret[0] else UserCompact, ret))

    def chat_keepalive(self, history_since: Optional[int] = None, since: Optional[int] = None) -> List[UserSilence]:
        """
        Request periodically to reset chat activity timeout. Also returns an updated list of recent silences.

        Requires OAuth, scope chat.read, and a user (authorization code grant or delegate scope)

        **Parameters**

        history_since: Optional[int]
            :class:`UserSilence`s after the specified id to return.
            This field is preferred and takes precedence over `since`.

        since: Optional[int]
            :class:`UserSilence`s after the specified :class:`ChatMessage`.message_id to return.

        **Returns**

        List[:class:`UserSilence`]
        """
        return list(
            map(
                UserSilence,
                self.http.make_request(Path.chat_keepalive(), history_since=history_since, since=since)["silences"],
            )
        )

    def create_new_pm(
        self, target_id: int, message: str, is_action: bool, uuid: Optional[str] = None
    ) -> CreateNewPmResult:
        """
        This endpoint allows you to create a new PM channel.

        Requires OAuth, scope chat.write, and a user (authorization code grant or delegate scope)

        **Parameters**

        target_id: int
            user_id of user to start PM with

        message: str
            message to send

        is_action: bool
            whether the message is an action

        uuid: Optional[str]
            client-side message identifier which will be sent back in response and websocket json.

        **Returns**

        :class:`CreateNewPmResult`
        """
        data = {"target_id": target_id, "message": message, "is_action": is_action}
        if uuid is not None:
            data["uuid"] = uuid
        resp = self.http.make_request(Path.create_new_pm(), files=create_multipart_formdata(data))
        return CreateNewPmResult(
            ChatChannel(resp["channel"]),
            ChatMessage(resp["message"]),
            resp.get("new_channel_id"),
        )

    def get_channel_messages(
        self, channel_id: int, limit: Optional[int] = None, since: Optional[int] = None, until: Optional[int] = None
    ) -> List[ChatMessage]:
        """
        This endpoint returns the chat messages for a specific channel.

        Requires OAuth, scope chat.read, and a user (authorization code grant or delegate scope)

        **Parameters**

        channel_id: int
            The ID of the channel to retrieve messages for

        limit: Optional[int]
            number of messages to return (max of 50)

        since: Optional[int]
            messages after the specified message id will be returned

        until: Optional[int]
            messages up to but not including the specified message id will be returned

        **Returns**

        List[:class:`ChatMessage`]
        """
        return list(
            map(
                ChatMessage,
                self.http.make_request(Path.get_channel_messages(channel_id), limit=limit, since=since, until=until),
            )
        )

    def send_message_to_channel(self, channel_id: int, message: str, is_action: bool) -> ChatMessage:
        """
        This endpoint sends and returns a chat messages for a specific channel.

        Requires OAuth, scope chat.write, and a user (authorization code grant or delegate scope)

        **Parameters**

        channel_id: int
            The id of the channel to send message to

        message: class:`str`
            message to send

        is_action: bool
            whether the message is an action
        """
        return ChatMessage(
            self.http.make_request(
                Path.send_message_to_channel(channel_id),
                files=create_multipart_formdata({"message": message, "is_action": is_action}),
            )
        )

    def join_channel(self, channel_id: int, user_id: int) -> ChatChannel:
        """
        This endpoint allows you to join a public or multiplayer channel.

        Requires OAuth, scope chat.write_manage, and a user (authorization code grant or delegate scope)

        **Parameters**

        channel_id: int
            The id of the channel to join

        user_id: int
            The id of the user to be joined

        **Returns**

        :class:`ChatChannel`
        """
        return ChatChannel(self.http.make_request(Path.join_channel(channel_id, user_id)))

    def leave_channel(self, channel_id: int, user_id: int) -> None:
        """
        This endpoint allows you to leave a public channel.

        Requires OAuth, scope chat.write_manage, and a user (authorization code grant or delegate scope)

        **Parameters**

        channel_id: int
            The id of the channel to join

        user_id: int
            The id of the user to be joined
        """
        self.http.make_request(Path.leave_channel(channel_id, user_id))

    def mark_channel_as_read(self, channel_id: int, message_id: int) -> None:
        """
        This endpoint marks the channel as having being read up to the given `message_id`.

        Requires OAuth, scope chat.read, and a user (authorization code grant or delegate scope)

        **Parameters**

        channel_id: int
            id of the channel to mark as read

        message_id: int
            most recent message to mark as read up to
        """
        self.http.make_request(Path.mark_channel_read(channel_id, message_id))

    def get_channel_list(self) -> List[ChatChannel]:
        """
        This endpoint returns a list of all joinable public channels.

        Requires OAuth, scope chat.read, and a user (authorization code grant or delegate scope)

        **Returns**

        List[:class:`ChatChannel`]
        """
        return list(map(ChatChannel, self.http.make_request(Path.get_channel_list())))

    def create_pm_channel(self, user_id: int) -> ChatChannel:
        """
        Create PM channel with another user. Rejoins if it already exists.

        Requires OAuth, scope chat.write_manage, and a user (authorization code grant or delegate scope)

        **Parameters**

        user_id: int
            id of the user to open a PM channel with

        **Returns**

        :class:`ChatChannel` with `recent_message`
        """
        return ChatChannel(
            self.http.make_request(
                Path.create_channel(),
                files=create_multipart_formdata({"target_id": user_id, "type": ChatChannelType.PM.value}),
            )
        )

    def create_announcement_channel(
        self, user_ids: List[int], message: str, name: str, description: str
    ) -> ChatChannel:
        """
        Creates a new announcement channel.

        Requires OAuth, scope chat.write_manage, and a user (authorization code grant or delegate scope)

        **Parameters**

        target_ids: :class:`List[int]`
            users to include in the channel

        message: str
            message to send with the announcement

        name: str
            the channel name

        description: str
            the channel description

        **Returns**

        :class:`ChatChannel` with `recent_message`
        """
        return ChatChannel(
            self.http.make_request(
                Path.create_channel(),
                files=create_multipart_formdata(
                    {
                        "target_ids": user_ids,
                        "message": message,
                        "channel": {"name": name, "description": description},
                    }
                ),
            )
        )

    def get_channel(self, channel_id: int) -> GetChannelResult:
        """
        Get a channel by id

        Requires OAuth, scope chat.read, and a user (authorization code grant or delegate scope)

        **Parameters**

        channel_id: int
            id of the channel to get

        **Returns**

        :class:`GetChannelResult`
        """
        ret = self.http.make_request(Path.get_channel(channel_id))
        return GetChannelResult(ChatChannel(ret["channel"]), list(map(UserCompact, ret["users"])))

    def get_all_scores(
        self, ruleset: Optional[Union[GameModeStr, str]] = None, cursor: Optional[str] = None
    ) -> GetAllScoresResult:
        """
        Returns all scores (that are passes), up to 1000. Result is sorted oldest to recent.
        Can use the cursor in the response in the next call to this function.

        **Parameters**

        ruleset: Optional[Union[:class:`GameModeStr`, str]]

        cursor: Optional[str]

        **Returns**

        List[:class:`SoloScore`]
        """
        ruleset = parse_enum_args(ruleset)

        ret = self.http.make_request(Path.get_all_scores(), ruleset=ruleset, cursor_string=cursor)
        return GetAllScoresResult(
            [get_score_object(score, self.http.api_version) for score in ret["scores"]], ret["cursor_string"]
        )

    def get_forums(self) -> GetForumsResult:
        """
        Returns top-level forums and their sub-forums (max 2 deep).

        **Returns**

        :class:`GetForumsResult`
        """
        ret = self.http.make_request(Path.get_forums())
        return GetForumsResult(list(map(Forum, ret["forums"])))

    def get_forum(self, forum_id: int) -> GetForumResult:
        """
        Get a forum, its recent topics, and pinned topics.

        **Parameters**

        forum_id: int

        **Returns**

        :class:`GetForumResult`
        """
        ret = self.http.make_request(Path.get_forum(forum_id))
        return GetForumResult(
            Forum(ret["forum"]),
            list(map(ForumTopic, ret["topics"])),
            list(map(ForumTopic, ret["pinned_topics"])),
        )

    def get_forum_topics(
        self,
        forum_id: Optional[int] = None,
        cursor: Optional[str] = None,
        sort: Optional[str] = None,
        limit: Optional[int] = None,
    ) -> GetForumTopicsResult:
        """
        Get a list of topics, optionally under a specified forum.

        **Parameters**

        forum_id: Optional[int]
            Id of forum to get topics from.
            Will retrieve topics from all forums if not specified.

        cursor: Optional[str]
            Pass cursor value of last call to continue

        sort: Optional[str]
            'new' (default) or 'old'

        limit: Optional[int]
            50 at most and by default

        **Returns**

        :class:`GetForumTopicsResult`
        """
        ret = self.http.make_request(
            Path.get_forum_topics(), forum_id=forum_id, cursor_string=cursor, sort=sort, limit=limit
        )
        return GetForumTopicsResult(
            list(map(ForumTopic, ret["topics"])),
            ret["cursor_string"],
        )



from .beatmap import *
from .score import *
from .discussion import *
from .build import *
from .chat import *
from .comment import *
from .event import *
from .forum import *
from .group import *
from .multiplayer import *
from .notification import *
from .ranking import *
from .user import *
from .wiki import *
from .kudosu import *
from .news import *
from .beatmapset_event import *
from .match import *
from .seasonal_background import *
from .current_user_attributes import *
from .achievement import *



from typing import Optional, List, TYPE_CHECKING

from ..util import prettify, get_optional, get_optional_list, get_required, fromisoformat

if TYPE_CHECKING:
    from datetime import datetime


__all__ = ("Build", "Versions", "UpdateStream", "ChangelogEntry", "GithubUser")


class Build:
    """
    **Attributes**

    created_at: :py:class:`datetime.datetime`

    display_version: str

    id: int

    update_stream: Optional[:class:`UpdateStream`]

    users: int

    version Optional[str]

    changelog_entries: Optional[List[:class:`ChangelogEntry`]]
        If the build has no changelog entries, a placeholder is generated.

    versions: Optional[:class:`Versions`]
    """

    __slots__ = (
        "created_at",
        "display_version",
        "id",
        "update_stream",
        "users",
        "version",
        "changelog_entries",
        "versions",
    )

    def __init__(self, data):
        self.created_at: datetime = fromisoformat(get_required(data, "created_at"))
        self.display_version: str = get_required(data, "display_version")
        self.id: int = get_required(data, "id")
        self.update_stream: Optional[UpdateStream] = get_optional(data, "update_stream", UpdateStream)
        self.users: int = get_required(data, "users")
        self.version: Optional[str] = get_required(data, "version")
        self.changelog_entries: Optional[List[ChangelogEntry]] = get_optional_list(
            data, "changelog_entries", ChangelogEntry
        )
        self.versions: Optional[Versions] = get_optional(data, "versions", Versions)

    def __repr__(self):
        return prettify(self, "id", "version")


class Versions:
    """
    **Optional Attributes**

    next: Optional[:class:`Build`]
        May be null if there is not a next build.

    previous: Optional[:class:`Build`]
        May be null if there is not a previous build.
    """

    __slots__ = ("next", "previous")

    def __init__(self, data):
        self.next: Optional[Build] = get_optional(data, "next", Build)
        self.previous: Optional[Build] = get_optional(data, "previous", Build)

    def __repr__(self):
        fields = [getattr(self, slot) for slot in self.__slots__ if getattr(self, slot, None)]
        return prettify(self, *fields)


class UpdateStream:
    """
    **Attributes**

    display_name: Optional[str]

    id: int

    is_featured: bool

    name: str

    latest_build: Optional[:class:`Build`]

    user_count: Optional[int]
    """

    __slots__ = (
        "display_name",
        "id",
        "is_featured",
        "name",
        "latest_build",
        "user_count",
    )

    def __init__(self, data):
        self.display_name: Optional[str] = get_required(data, "display_name")
        self.id: int = get_required(data, "id")
        self.is_featured: bool = get_required(data, "is_featured")
        self.name: str = get_required(data, "name")
        self.latest_build: Optional[Build] = get_optional(data, "latest_build", Build)
        self.user_count: Optional[int] = data.get("user_count")

    def __repr__(self):
        return prettify(self, "name")


class ChangelogEntry:
    """
    **Attributes**

    category: str

    created_at: Optional[:py:class:`datetime.datetime`]

    github_pull_request_id: Optional[int]

    github_url: Optional[str]

    id: Optional[int]

    major: bool

    repository: Optional[str]

    title: Optional[str]

    type: str

    url: Optional[str]

    github_user: Optional[:class:`GithubUser`]
        If the changelog entry has no GitHub user, a placeholder is generated.

    message: Optional[str]
        Entry message in Markdown format. Embedded HTML is allowed.

    message_html: Optional[str]
        Entry message in HTML format.
    """

    __slots__ = (
        "category",
        "created_at",
        "github_pull_request_id",
        "github_url",
        "id",
        "major",
        "repository",
        "title",
        "type",
        "url",
        "github_user",
        "message",
        "message_html",
    )

    def __init__(self, data):
        self.category: str = get_required(data, "category")
        self.created_at: Optional[datetime] = get_optional(data, "created_at", fromisoformat)
        self.github_pull_request_id: Optional[int] = get_required(data, "github_pull_request_id")
        self.github_url: Optional[str] = get_required(data, "github_url")
        self.id: Optional[int] = get_required(data, "id")
        self.major: bool = get_required(data, "major")
        self.repository: Optional[str] = get_required(data, "repository")
        self.title: Optional[str] = get_required(data, "title")
        self.type: str = get_required(data, "type")
        self.url: Optional[str] = get_required(data, "url")

        self.github_user: Optional[GithubUser] = get_optional(data, "github_user", GithubUser)
        self.message: Optional[str] = data.get("message")
        self.message_html: Optional[str] = data.get("message_html")

    def __repr__(self):
        return prettify(self, "title", "major", "created_at")


class GithubUser:
    """
    **Attributes**

    display_name: str

    github_url: Optional[str]

    id: Optional[int]

    osu_username: Optional[str]

    user_id: Optional[int]

    user_url: Optional[str]
    """

    __slots__ = (
        "display_name",
        "github_url",
        "id",
        "osu_username",
        "user_id",
        "user_url",
    )

    def __init__(self, data):
        self.display_name: str = get_required(data, "display_name")
        self.github_url: Optional[str] = get_required(data, "github_url")
        self.id: Optional[int] = get_required(data, "id")
        self.osu_username: Optional[str] = get_required(data, "osu_username")
        self.user_id: Optional[int] = get_required(data, "user_id")
        self.user_url: Optional[str] = get_required(data, "user_url")

    def __repr__(self):
        return prettify(self, "display_name")



from typing import Optional

from ..util import prettify, get_optional, get_required
from ..enums import GameModeStr


__all__ = ("Achievement",)


class Achievement:
    """
    **Attributes**

    icon_url: str

    id: int

    name: str

    grouping: str

    ordering: int

    slug: str

    description: str

    mode: Optional[:class:`GameModeStr`]

    instructions: Optional[str]
    """

    __slots__ = (
        "icon_url",
        "id",
        "name",
        "grouping",
        "ordering",
        "slug",
        "description",
        "mode",
        "instructions",
    )

    def __init__(self, data):
        self.icon_url: str = get_required(data, "icon_url")
        self.id: int = get_required(data, "id")
        self.name: str = get_required(data, "name")
        self.grouping: str = get_required(data, "grouping")
        self.ordering: int = get_required(data, "ordering")
        self.slug: str = get_required(data, "slug")
        self.description: str = get_required(data, "description")
        self.mode: GameModeStr = get_optional(data, "mode", GameModeStr)
        self.instructions: Optional[str] = get_required(data, "instructions")

    def __repr__(self):
        return prettify(self, "name", "description")



__all__ = ("ScopeException", "RequestException")


class ScopeException(Exception):
    """Raised when there was an exception involving scopes"""


class RequestException(Exception):
    """Raised when there was an exception involving requests to the api"""
